###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_PIC_16_LE_16_PIC_16F(ArchPcode):
    name = 'PIC-16:LE:16:PIC-16F'
    pcode_arch = 'PIC-16:LE:16:PIC-16F'
    description = 'PIC-16F(L)XXX'
    bits = 16
    ip_offset = 0x0
    sp_offset = 0x2
    bp_offset = sp_offset
    instruction_endness = Endness.LE
    register_list = [
        Register('indf0', 1, 0x0),
        Register('indf1', 1, 0x1),
        Register('pcl', 1, 0x2),
        Register('status', 1, 0x3),
        Register('fsr0', 2, 0x4),
        Register('fsr0l', 1, 0x4),
        Register('fsr0h', 1, 0x5),
        Register('fsr1', 2, 0x6),
        Register('fsr1l', 1, 0x6),
        Register('fsr1h', 1, 0x7),
        Register('bsr', 1, 0x8),
        Register('wreg', 1, 0x9),
        Register('pclath', 1, 0xa),
        Register('intcon', 1, 0xb),
        Register('pc', 2, 0x0, alias_names=('ip',)),
        Register('stkptr', 1, 0x2),
        Register('w', 1, 0x3),
        Register('skipnext', 1, 0x4),
        Register('irp', 1, 0x7),
        Register('rp', 1, 0x8),
        Register('contextreg', 4, 0x100)
    ]

register_arch(['pic-16:le:16:pic-16f'], 16, Endness.LE, ArchPcode_PIC_16_LE_16_PIC_16F)



from angr.analyses.cfg.cfg_job_base import BlockID, CFGJobBase


class JobInfo:
    """
    Stores information of each job.
    """
    def __init__(self, key: BlockID, job: CFGJobBase):
        self.key = key
        self.jobs = [(job, '')]

        self.narrowing_count = 0  # not used

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, o):
        return type(self) == type(o) and \
               self.key == o.key

    def __repr__(self):
        s = "<JobInfo %s>" % (str(self.key))
        return s

    @property
    def job(self) -> CFGJobBase:
        """
        Get the latest available job.

        :return: The latest available job.
        """

        job, _ = self.jobs[-1]
        return job

    @property
    def merged_jobs(self):
        for job, job_type in self.jobs:
            if job_type == 'merged':
                yield job

    @property
    def widened_jobs(self):
        for job, job_type in self.jobs:
            if job_type == 'widened':
                yield job

    def add_job(self, job, merged=False, widened=False):
        """
        Appended a new job to this JobInfo node.
        :param job: The new job to append.
        :param bool merged: Whether it is a merged job or not.
        :param bool widened: Whether it is a widened job or not.
        """

        job_type = ''
        if merged:
            job_type = 'merged'
        elif widened:
            job_type = 'widened'
        self.jobs.append((job, job_type))



from typing import TypeVar, Generic, List, Collection, Optional, Iterator, Set, Dict

from ....misc.ux import deprecated
from ....utils.algo import binary_insert

NodeType = TypeVar("NodeType")


class GraphVisitor(Generic[NodeType]):
    """
    A graph visitor takes a node in the graph and returns its successors. Typically it visits a control flow graph, and
    returns successors of a CFGNode each time. This is the base class of all graph visitors.
    """

    __slots__ = ('_sorted_nodes', '_nodes_set', '_node_to_index', '_reached_fixedpoint', )

    def __init__(self):
        self._sorted_nodes: List[NodeType] = [ ]
        self._nodes_set: Set[NodeType] = set()
        self._node_to_index: Dict[NodeType, int] = { }
        self._reached_fixedpoint: Set[NodeType] = set()

    #
    # Interfaces
    #

    def successors(self, node: NodeType) -> List[NodeType]:
        """
        Get successors of a node. The node should be in the graph.

        :param node: The node to work with.
        :return:     A list of successors.
        :rtype:      list
        """

        raise NotImplementedError()

    def predecessors(self, node: NodeType) -> List[NodeType]:
        """
        Get predecessors of a node. The node should be in the graph.

        :param node: The node to work with.
        :return:     A list of predecessors.
        :rtype:      list
        """

        raise NotImplementedError()

    def sort_nodes(self, nodes: Optional[Collection[NodeType]] = None) -> List[NodeType]:
        """
        Get a list of all nodes sorted in an optimal traversal order.

        :param iterable nodes: A collection of nodes to sort. If none, all nodes in the graph will be used to sort.
        :return:               A list of sorted nodes.
        :rtype:                list
        """

        raise NotImplementedError()

    #
    # Public methods
    #

    def nodes(self) -> Iterator[NodeType]:
        """
        Return an iterator of nodes following an optimal traversal order.

        :return:
        """

        return iter(
            self.sort_nodes()
        )

    @deprecated(replacement='nodes')
    def nodes_iter(self):
        """
        (Deprecated) Return an iterator of nodes following an optimal traversal order. Will be removed in the future.
        """
        return self.nodes()

    # Traversal

    def reset(self):
        """
        Reset the internal node traversal state. Must be called prior to visiting future nodes.

        :return: None
        """

        self._sorted_nodes.clear()
        self._nodes_set.clear()
        self._node_to_index.clear()
        self._reached_fixedpoint.clear()

        for i, n in enumerate(self.sort_nodes()):
            self._node_to_index[n] = i
            binary_insert(self._sorted_nodes, n, lambda elem: self._node_to_index[elem])
            self._nodes_set.add(n)

    def next_node(self) -> Optional[NodeType]:
        """
        Get the next node to visit.

        :return: A node in the graph.
        """

        if not self._sorted_nodes:
            return None

        node = self._sorted_nodes.pop(0)
        self._nodes_set.remove(node)
        return node

    def all_successors(self, node: NodeType, skip_reached_fixedpoint=False) -> Set[NodeType]:
        """
        Returns all successors to the specific node.

        :param node: A node in the graph.
        :return:     A set of nodes that are all successors to the given node.
        :rtype:      set
        """

        successors = set()

        stack = [ node ]
        while stack:
            n = stack.pop()
            successors.add(n)
            stack.extend(succ for succ in self.successors(n) if
                         succ not in successors and
                            (not skip_reached_fixedpoint or succ not in self._reached_fixedpoint)
                         )

        return successors

    def revisit_successors(self, node: NodeType, include_self=True) -> None:
        """
        Revisit a node in the future. As a result, the successors to this node will be revisited as well.

        :param node: The node to revisit in the future.
        :return:     None
        """

        successors = self.successors(node) #, skip_reached_fixedpoint=True)

        if include_self:
            if node not in self._nodes_set:
                binary_insert(self._sorted_nodes, node, lambda elem: self._node_to_index[elem])
                self._nodes_set.add(node)

        for succ in successors:
            if succ not in self._nodes_set:
                binary_insert(self._sorted_nodes, succ, lambda elem: self._node_to_index[elem])
                self._nodes_set.add(succ)

    def revisit_node(self, node: NodeType) -> None:
        """
        Revisit a node in the future. Do not include its successors immediately.

        :param node:    The node to revisit in the future.
        :return:        None
        """

        if node not in self._nodes_set:
            binary_insert(self._sorted_nodes, node, lambda elem: self._node_to_index[elem])
            self._nodes_set.add(node)

    def reached_fixedpoint(self, node: NodeType) -> None:
        """
        Mark a node as reached fixed-point. This node as well as all its successors will not be visited in the future.

        :param node: The node to mark as reached fixed-point.
        :return:     None
        """

        self._reached_fixedpoint.add(node)



# pylint:disable=no-self-use,arguments-renamed,isinstance-second-argument-not-valid-type
from typing import Optional

import ailment
import claripy

from ..structurer_nodes import ConditionNode, CascadingConditionNode
from ..sequence_walker import SequenceWalker


class CascadingConditionTransformer(SequenceWalker):
    """
    Identifies and transforms `if { ... } else { if { ... } else { ... } }` to
    `if { ... } else if { ... } else if { ... }`.
    """
    def __init__(self, node):
        handlers = {
            ConditionNode: self._handle_Condition,
        }
        super().__init__(handlers)
        self.cascading_if_node: Optional[CascadingConditionNode] = None

        self.walk(node)

    def _negate_ailexpr(self, expr: ailment.Expression) -> ailment.Expression:
        return ailment.UnaryOp(None, 'Not', expr, **expr.tags)

    def _handle_Condition(self, cond_node: ConditionNode, **kwargs):
        if cond_node.false_node is not None \
                and isinstance(cond_node.false_node, (ConditionNode, CascadingConditionNode)) \
                and not isinstance(cond_node.true_node, (ConditionNode, CascadingConditionNode)) \
                and cond_node.true_node is not None:

            cond_0 = cond_node.condition
            node_0 = cond_node.true_node
            remaining_node = cond_node.false_node

        elif cond_node.true_node is not None \
                and isinstance(cond_node.true_node, (ConditionNode, CascadingConditionNode)) \
                and not isinstance(cond_node.false_node, (ConditionNode, CascadingConditionNode)) \
                and cond_node.false_node is not None:
            if isinstance(cond_node.condition, claripy.ast.Base):
                cond_0 = claripy.Not(cond_node.condition)
            else:
                cond_0 = self._negate_ailexpr(cond_node.condition)
            node_0 = cond_node.false_node
            remaining_node = cond_node.true_node

        else:
            return super()._handle_Condition(cond_node, **kwargs)

        # structure else_node
        if not isinstance(remaining_node, CascadingConditionNode):
            structured = self._handle_Condition(remaining_node)
            if structured is None:
                structured = remaining_node
        else:
            structured = remaining_node

        if isinstance(structured, ConditionNode):
            if structured.true_node is None and structured.false_node is not None:
                if isinstance(structured.condition, claripy.ast.Base):
                    negated_structured_condition = claripy.Not(structured.condition)
                else:
                    negated_structured_condition = self._negate_ailexpr(structured.condition)
                cond_and_nodes = [
                    (cond_0, node_0),
                    (negated_structured_condition, structured.false_node),
                ]
                else_node = None
            elif structured.true_node is not None and structured.false_node is None:
                cond_and_nodes = [
                    (cond_0, node_0),
                    (structured.condition, structured.true_node),
                ]
                else_node = None
            else:
                cond_and_nodes = [
                    (cond_0, node_0),
                    (structured.condition, structured.true_node),
                ]
                else_node = structured.false_node

        elif isinstance(structured, CascadingConditionNode):
            # merge two nodes
            cond_and_nodes = [
                                 (cond_0, node_0)
                             ] + structured.condition_and_nodes
            else_node = structured.else_node

        else:
            # unexpected!
            raise RuntimeError("Impossible happened")

        return CascadingConditionNode(cond_node.addr, cond_and_nodes, else_node=else_node)



import angr

"""
references:
https://elixir.bootlin.com/linux/latest/source/arch/arm/kernel/entry-armv.S
https://www.kernel.org/doc/Documentation/arm/kernel_user_helpers.txt
"""

class _kuser_helper_version(angr.SimProcedure):
    # pylint: disable=arguments-differ
    kuser_addr = 0xffff0ffc
    def run(self):
        # hardcoded version number extracted from QEMU
        self.state.regs.r0 = 0x884c0
        return

class _kuser_helper_get_tls(angr.SimProcedure):
    # pylint: disable=arguments-differ
    kuser_addr = 0xffff0fe0
    def run(self):
        self.state.regs.r0 = self.project.loader.tls.threads[0].user_thread_pointer
        return

class _kuser_cmpxchg(angr.SimProcedure):
    # pylint: disable=arguments-differ
    kuser_addr = 0xffff0fc0
    def run(self, oldval, newval, ptr):
        currentval = self.state.mem[ptr].dword.resolved
        success = currentval == oldval
        self.state.memory.store(ptr, newval, condition=success)

        # set flags
        retval = currentval - oldval
        self.state.regs.cc_op = 2
        self.state.regs.cc_dep1 = 0
        self.state.regs.cc_dep2 = retval
        self.state.regs.cc_ndep = 0
        return retval

class _kuser_memory_barrier(angr.SimProcedure):
    # pylint: disable=arguments-differ
    kuser_addr = 0xffff0fa0
    def run(self):
        pass

class _kuser_cmpxchg64(angr.SimProcedure):
    # pylint: disable=arguments-differ
    kuser_addr = 0xffff0f60
    def run(self):
        raise NotImplementedError(f"{self.__class__.__name__} is not implemented")



import logging
l = logging.getLogger(name=__name__)

import claripy
import functools

from .. import sim_options as o

#pylint:disable=unidiomatic-typecheck

_noneset = frozenset()

def _raw_ast(a):
    if type(a) is SimActionObject:
        return a.ast
    elif type(a) is dict:
        return { k:_raw_ast(a[k]) for k in a }
    elif type(a) in (tuple, list, set, frozenset):
        return type(a)((_raw_ast(b) for b in a))
    elif type(a) in (zip, filter, map):
        return (_raw_ast(i) for i in a)
    else:
        return a

def _all_objects(a):
    if type(a) is SimActionObject:
        yield a
    elif type(a) is dict:
        for b in a.values():
            for o in _all_objects(b):
                yield o
    elif type(a) is (tuple, list, set, frozenset):
        for b in a:
            for o in _all_objects(b):
                yield o

def ast_stripping_op(f, *args, **kwargs):
    new_args = _raw_ast(args)
    new_kwargs = _raw_ast(kwargs)
    return f(*new_args, **new_kwargs)

def ast_preserving_op(f, *args, **kwargs):
    tmp_deps = frozenset.union(_noneset, *(a.tmp_deps for a in _all_objects(args)))
    reg_deps = frozenset.union(_noneset, *(a.reg_deps for a in _all_objects(args)))

    a = ast_stripping_op(f, *args, **kwargs)
    if isinstance(a, claripy.ast.Base):
        return SimActionObject(a, reg_deps=reg_deps, tmp_deps=tmp_deps)
    else:
        return a

def ast_stripping_decorator(f):
    @functools.wraps(f)
    def ast_stripper(*args, **kwargs):
        new_args = _raw_ast(args)
        new_kwargs = _raw_ast(kwargs)
        return f(*new_args, **new_kwargs)
    return ast_stripper


class SimActionObject:
    """
    A SimActionObject tracks an AST and its dependencies.
    """
    def __init__(self, ast, reg_deps=None, tmp_deps=None, deps=None, state=None):
        if type(ast) is SimActionObject:
            raise SimActionError("SimActionObject inception!!!")
        self.ast = ast
        if deps is not None:
            if len(deps) == 0 or (state is not None and o.ACTION_DEPS not in state.options):
                self.reg_deps = _noneset
                self.tmp_deps = _noneset
            else:
                self.reg_deps = frozenset.union(*[r.reg_deps for r in deps if type(r) in (sim_action.SimActionData, sim_action.SimActionOperation)])
                self.tmp_deps = frozenset.union(*[r.tmp_deps for r in deps if type(r) in (sim_action.SimActionData, sim_action.SimActionOperation)])
        else:
            self.reg_deps = _noneset if reg_deps is None else reg_deps
            self.tmp_deps = _noneset if tmp_deps is None else tmp_deps

    def __repr__(self):
        return '<SAO {}>'.format(self.ast)

    def __getstate__(self):
        return self.ast, self.reg_deps, self.tmp_deps

    def __setstate__(self, data):
        self.ast, self.reg_deps, self.tmp_deps = data

    def _preserving_unbound(self, f, *args, **kwargs):
        return ast_preserving_op(f, *((self,) + tuple(args)), **kwargs)

    def _preserving_bound(self, f, *args, **kwargs): #pylint:disable=no-self-use
        return ast_preserving_op(f, *args, **kwargs)

    def __getattr__(self, attr):
        if attr == '__slots__':
            raise AttributeError("not forwarding __slots__ to AST")

        f = getattr(self.ast, attr)
        if hasattr(f, '__call__'):
            return functools.partial(self._preserving_bound, f)
        elif isinstance(f, claripy.ast.Base):
            return SimActionObject(f, reg_deps=self.reg_deps, tmp_deps=self.tmp_deps)
        else:
            return f

    def __len__(self):
        return len(self.ast)

    def to_claripy(self):
        return self.ast

    def copy(self):
        return SimActionObject(self.ast, self.reg_deps, self.tmp_deps)

#
# Overload the operators
#

def _operator(cls, op_name):
    def wrapper(self, *args, **kwargs):
        return self._preserving_unbound(getattr(self.ast.__class__, op_name), *args, **kwargs)
    wrapper.__name__ = op_name
    setattr(cls, op_name, wrapper)

def make_methods():
    for name in claripy.operations.expression_operations | { '__getitem__' } :
        _operator(SimActionObject, name)
make_methods()

from ..errors import SimActionError
from . import sim_action



import logging
from itertools import count
from typing import Dict, Optional, Generator, Union, TYPE_CHECKING, Tuple, Iterable

import claripy
from claripy.ast import Bool, Bits, BV
from claripy.vsa import StridedInterval, ValueSet, RegionAnnotation

from ....sim_options import (HYBRID_SOLVER, APPROXIMATE_FIRST, AVOID_MULTIVALUED_READS, CONSERVATIVE_READ_STRATEGY,
    KEEP_MEMORY_READS_DISCRETE, CONSERVATIVE_WRITE_STRATEGY)
from ....state_plugins.sim_action_object import _raw_ast
from ....errors import SimMemoryError, SimAbstractMemoryError
from .. import MemoryMixin
from .region_data import AddressWrapper, RegionMap
from .abstract_address_descriptor import AbstractAddressDescriptor

if TYPE_CHECKING:
    from ....sim_state import SimState
    from .. import RegionedMemory


_l = logging.getLogger(name=__name__)

invalid_read_ctr = count()


class RegionedMemoryMixin(MemoryMixin):
    """
    Regioned memory. This mixin manages multiple memory regions. Each address is represented as a tuple of (region ID,
    offset into the region), which is called a regioned address.

    Converting absolute addresses into regioned addresses: We map an absolute address to a region by looking up which
    region this address belongs to in the region map. Currently this is only enabled for stack. Heap support has not
    landed yet.

    When start analyzing a function, the user should call set_stack_address_mapping() to create a new region mapping.
    Likewise, when exiting from a function, the user should cancel the previous mapping by calling
    unset_stack_address_mapping().
    """

    def __init__(self, write_targets_limit: int=2048, read_targets_limit: int=4096,
                 stack_region_map: Optional[RegionMap]=None,
                 generic_region_map: Optional[RegionMap]=None,
                 stack_size: int=65536,
                 cle_memory_backer: Optional=None,
                 dict_memory_backer: Optional[Dict]=None,
                 regioned_memory_cls: Optional[type]=None,
                 **kwargs):
        super().__init__(**kwargs)

        if regioned_memory_cls is None:
            # delayed import
            from .. import RegionedMemory
            regioned_memory_cls = RegionedMemory

        self._regioned_memory_cls = regioned_memory_cls
        self._regions: Dict[str,regioned_memory_cls] = { }

        self._cle_memory_backer = cle_memory_backer
        self._dict_memory_backer = dict_memory_backer
        self._stack_size: int = stack_size
        self._stack_region_map: Optional[RegionMap] = stack_region_map \
            if stack_region_map is not None else RegionMap(True)
        self._generic_region_map: Optional[RegionMap] = generic_region_map \
            if generic_region_map is not None else RegionMap(False)

        self._write_targets_limit = write_targets_limit
        self._read_targets_limit = read_targets_limit

    @MemoryMixin.memo
    def copy(self, memo):
        o: 'RegionedMemoryMixin' = super().copy(memo)
        o._write_targets_limit = self._write_targets_limit
        o._read_targets_limit = self._read_targets_limit
        o._stack_size = self._stack_size
        o._endness = self.endness
        o._stack_region_map = self._stack_region_map
        o._generic_region_map = self._generic_region_map
        o._cle_memory_backer = self._cle_memory_backer
        o._dict_memory_backer = self._dict_memory_backer
        o._regioned_memory_cls = self._regioned_memory_cls

        o._regions = {}
        for region_id, region in self._regions.items():
            o._regions[region_id] = region.copy(memo)

        return o

    def load(self, addr, size: Optional[Union[BV,int]]=None, endness=None, condition: Optional[Bool]=None, **kwargs):

        if isinstance(size, BV) and isinstance(size._model_vsa, ValueSet):
            _l.critical('load(): size %s is a ValueSet. Something is wrong.', size)
            if self.state.scratch.ins_addr is not None:
                var_name = 'invalid_read_%d_%#x' % (
                    next(invalid_read_ctr),
                    self.state.scratch.ins_addr
                )
            else:
                var_name = 'invalid_read_%d_None' % next(invalid_read_ctr)

            return self.state.solver.Unconstrained(var_name, self.state.arch.bits)

        val = None
        regioned_addrs_desc = self._normalize_address(addr, condition=condition)

        if (regioned_addrs_desc.cardinality > 1 and AVOID_MULTIVALUED_READS in self.state.options) or \
                (regioned_addrs_desc.cardinality >= self._read_targets_limit and CONSERVATIVE_READ_STRATEGY in self.state.options):
            val = self.state.solver.Unconstrained('unconstrained_read', size * self.state.arch.byte_width)
            return val

        gen = self._concretize_address_descriptor(regioned_addrs_desc, addr, is_write=False)
        for aw in gen:
            new_val = self._region_load(aw.address, size, aw.region,
                                        endness=endness,
                                        related_function_addr=aw.function_address,
                                        **kwargs,
                                        )

            if val is None:
                if KEEP_MEMORY_READS_DISCRETE in self.state.options:
                    val = self.state.solver.DSIS(to_conv=new_val, max_card=100000)
                else:
                    val = new_val
            else:
                val = val.union(new_val)

        if val is None:
            # address_wrappers is empty - we cannot concretize the address in static mode.
            # ensure val is not None
            val = self.state.solver.Unconstrained('invalid_read_%d_%d' % (next(invalid_read_ctr), size),
                                                  size * self.state.arch.byte_width)

        return val

    def store(self, addr, data, size: Optional[int]=None, endness=None, **kwargs):
        regioned_addrs_desc = self._normalize_address(addr)
        if regioned_addrs_desc.cardinality >= self._write_targets_limit and CONSERVATIVE_WRITE_STRATEGY in self.state.options:
            return

        gen = self._concretize_address_descriptor(regioned_addrs_desc, addr, is_write=True)
        for aw in gen:
            self._region_store(aw.address, data, aw.region, endness, related_function_addr=aw.function_address)

    def merge(self, others: Iterable['RegionedMemoryMixin'], merge_conditions, common_ancestor=None) -> bool:
        r = False
        for o in others:
            for region_id, region in o._regions.items():
                if region_id in self._regions:
                    r |= self._regions[region_id].merge([region], merge_conditions, common_ancestor=common_ancestor)
                else:
                    self._regions[region_id] = region
                    r = True
        return r

    def find(self, addr: Union[int,Bits], data, max_search, **kwargs):
        # FIXME: Attempt find() on more than one region

        gen = self._normalize_address_type(addr, self.state.arch.bits)

        for region, si in gen:
            si = claripy.SI(to_conv=si)
            r, s, i = self._regions[region].find(si, data, max_search, **kwargs)
            # Post-process r so that it's still a ValueSet
            region_base_addr = self._region_base(region)
            r = self.state.solver.ValueSet(r.size(), region, region_base_addr, r._model_vsa)
            return r, s, i

    def set_state(self, state):
        for region in self._regions.values():
            region.set_state(state)
        super().set_state(state)

    def replace_all(self, old: claripy.ast.BV, new: claripy.ast.BV):
        for region in self._regions.values():
            region.replace_all(old, new)

    #
    # Region management
    #

    def set_stack_address_mapping(self, absolute_address: int, region_id: str,
                                  related_function_address: Optional[int]=None):
        """
        Create a new mapping between an absolute address (which is the base address of a specific stack frame) and a
        region ID.

        :param absolute_address: The absolute memory address.
        :param region_id: The region ID.
        :param related_function_address: Related function address.
        """
        if self._stack_region_map is None:
            raise SimMemoryError('Stack region map is not initialized.')
        self._stack_region_map.map(absolute_address, region_id, related_function_address=related_function_address)

    def unset_stack_address_mapping(self, absolute_address: int):
        """
        Remove a stack mapping.

        :param absolute_address: An absolute memory address that is the base address of the stack frame to destroy.
        """
        if self._stack_region_map is None:
            raise SimMemoryError('Stack region map is not initialized.')
        self._stack_region_map.unmap_by_address(absolute_address)

    def stack_id(self, function_address: int) -> str:
        """
        Return a memory region ID for a function. If the default region ID exists in the region mapping, an integer
        will appended to the region name. In this way we can handle recursive function calls, or a function that
        appears more than once in the call frame.

        This also means that `stack_id()` should only be called when creating a new stack frame for a function. You are
        not supposed to call this function every time you want to map a function address to a stack ID.

        :param function_address: Address of the function.
        :return:                ID of the new memory region.
        """
        region_id = 'stack_%#x' % function_address

        # deduplication
        region_ids = self._stack_region_map.region_ids
        if region_id not in region_ids:
            return region_id
        else:
            for i in range(0, 2000):
                new_region_id = region_id + '_%d' % i
                if new_region_id not in region_ids:
                    return new_region_id
            raise SimMemoryError('Cannot allocate region ID for function %#08x - recursion too deep' % function_address)

    def set_stack_size(self, size: int):
        self._stack_size = size

    def _create_region(self, key: str, state: 'SimState', related_function_addr: int, endness,
                       cle_memory_backer: Optional=None, dict_memory_backer: Optional[Dict]=None):
        """
        Create a new MemoryRegion with the region key specified, and store it to self._regions.

        :param key: a string which is the region key
        :param state: the SimState instance
        :param is_stack: Whether this memory region is on stack. True/False
        :param related_function_addr: Which function first creates this memory region. Just for reference.
        :param endness: The endianness.
        :param backer_dict: The memory backer object.
        :return: None
        """
        self._regions[key] = self._regioned_memory_cls(
            memory_id=key,
            related_function_addr=related_function_addr,
            endness=endness,
            cle_memory_backer=cle_memory_backer,
            dict_memory_backer=dict_memory_backer,
        )
        self._regions[key].set_state(state)

    def _region_base(self, region: str) -> int:
        """
        Get the base address of a memory region.

        :param region:  ID of the memory region
        :return:        Address of the memory region
        """

        if region == 'global':
            base_addr = 0
        elif region.startswith('stack_'):
            base_addr = self._stack_region_map.absolutize(region, 0)
        else:
            base_addr = self._generic_region_map.absolutize(region, 0)

        return base_addr

    def _region_load(self, addr, size, key: str, related_function_addr=None, **kwargs):
        bbl_addr, stmt_id, ins_addr = self.state.scratch.bbl_addr, self.state.scratch.stmt_idx, self.state.scratch.ins_addr

        if key not in self._regions:
            self._create_region(key, self.state, related_function_addr, self.endness,
                                cle_memory_backer=self._cle_memory_backer.get(key, None) if self._cle_memory_backer is not None else None,
                                dict_memory_backer=self._dict_memory_backer.get(key, None) if self._dict_memory_backer is not None else None,
                                )

        return self._regions[key].load(addr, size, bbl_addr, stmt_id, ins_addr, **kwargs)

    def _region_store(self, addr, data, key: str, endness, related_function_addr: Optional[int]=None, **kwargs):
        if key not in self._regions:
            self._create_region(key, self.state, related_function_addr, self.endness,
                                cle_memory_backer=self._cle_memory_backer.get(key, None) if self._cle_memory_backer is not None else None,
                                dict_memory_backer=self._dict_memory_backer.get(key, None) if self._dict_memory_backer is not None else None,
                                )

        self._regions[key].store(addr, data,
                                 self.state.scratch.bbl_addr,
                                 self.state.scratch.stmt_idx,
                                 self.state.scratch.ins_addr,
                                 endness=endness,
                                 **kwargs
                                 )

    #
    # Address concretization and conversion
    #

    def _concretize_address_descriptor(self, desc: AbstractAddressDescriptor, original_addr: claripy.ast.Bits,
                                       is_write: bool=False,
                                       target_region: Optional[str]=None) -> Generator[AddressWrapper,None,None]:
        raise NotImplementedError()

    def _normalize_address(self, addr: claripy.ast.Bits, condition=None) -> AbstractAddressDescriptor:
        """
        Translate an address into a series of internal representation of addresses that can be used to address in
        individual regions.

        :param addr:
        :param is_write:
        :param convert_to_valueset:
        :param target_region:
        :return:
        """

        if type(addr) is not int:
            for constraint in self.state.solver.constraints:
                if getattr(addr, 'variables', set()) & constraint.variables:
                    addr = self._apply_condition_to_symbolic_addr(addr, constraint)

        # Apply the condition if necessary
        if condition is not None:
            addr = self._apply_condition_to_symbolic_addr(addr, condition)

        addr_with_regions = self._normalize_address_type(addr, self.state.arch.bits)

        desc = AbstractAddressDescriptor()
        for region, addr_si in addr_with_regions:
            desc.add_regioned_address(region, addr_si)
        return desc

    def _normalize_address_core(self, region_id: str, relative_address: int,
                                target_region: Optional[str]=None) -> AddressWrapper:
        """
        If this is a stack address, we convert it to a correct region and address

        :param region_id: a string indicating which region the address is relative to
        :param relative_address: an address that is relative to the region parameter
        :param target_region: the ideal target region that address is normalized to. None means picking the best fit.
        :return: an AddressWrapper object
        """
        if self._stack_region_map.is_empty and self._generic_region_map.is_empty:
            # We don't have any mapped region right now
            return AddressWrapper(region_id, 0, relative_address, False, None)

        # We wanna convert this address to an absolute address first
        if region_id.startswith('stack_'):
            absolute_address = self._stack_region_map.absolutize(region_id, relative_address)

        else:
            absolute_address = self._generic_region_map.absolutize(region_id, relative_address)

        stack_base = self._stack_region_map.stack_base

        if stack_base - self._stack_size < relative_address <= stack_base and \
                (target_region is not None and target_region.startswith('stack_')):
            # The absolute address seems to be in the stack region.
            # Map it to stack
            new_region_id, new_relative_address, related_function_addr = self._stack_region_map.relativize(
                absolute_address,
                target_region_id=target_region
            )

            return AddressWrapper(new_region_id, self._region_base(new_region_id), new_relative_address, True,
                                  related_function_addr
                                  )

        else:
            new_region_id, new_relative_address, related_function_addr = self._generic_region_map.relativize(
                absolute_address,
                target_region_id=target_region
            )

            return AddressWrapper(new_region_id, self._region_base(new_region_id), new_relative_address, False, None)

    def _apply_condition_to_symbolic_addr(self, addr, condition):
        _, converted = self.state.solver.constraint_to_si(condition)
        for original_expr, constrained_expr in converted:
            addr = addr.replace(original_expr, constrained_expr)
        return addr

    @staticmethod
    def _normalize_address_type(addr: Union[int,Bits], bits) -> Generator[Tuple[str,StridedInterval],None,None]:
        """
        Convert address of different types to a list of mapping between region IDs and offsets (strided intervals).

        :param addr: Address to convert
        :return: A list of mapping between region IDs and offsets.
        :rtype: dict
        """

        if isinstance(addr, int):
            addr_e = claripy.BVV(addr, bits)
        else:
            addr_e = _raw_ast(addr)

        if isinstance(addr_e, (claripy.bv.BVV, claripy.vsa.StridedInterval, claripy.vsa.ValueSet)):
            raise SimMemoryError('_normalize_address_type() does not take claripy models.')

        if isinstance(addr_e, claripy.ast.Base):
            if not isinstance(addr_e._model_vsa, ValueSet):
                # Convert it to a ValueSet first by annotating it
                addr_e = addr_e.annotate(RegionAnnotation('global', 0, addr_e._model_vsa))

            for region, offset in addr_e._model_vsa.items():
                yield region, offset

        else:
            raise SimAbstractMemoryError('Unsupported address type %s' % type(addr_e))



# pylint:disable=arguments-differ
from typing import Optional, TYPE_CHECKING
import logging

import claripy
import ailment

from ...calling_conventions import SimRegArg
from ...sim_type import SimTypeFunction, SimTypeBottom
from ...engines.light import SimEngineLightAILMixin
from ..typehoon import typeconsts, typevars
from ..typehoon.lifter import TypeLifter
from .engine_base import SimEngineVRBase, RichR

if TYPE_CHECKING:
    from .variable_recovery_fast import VariableRecoveryFastState


l = logging.getLogger(name=__name__)


class SimEngineVRAIL(
    SimEngineLightAILMixin,
    SimEngineVRBase,
):
    state: 'VariableRecoveryFastState'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._reference_spoffset: bool = False

    # Statement handlers

    def _ail_handle_Assignment(self, stmt):
        dst_type = type(stmt.dst)

        if dst_type is ailment.Expr.Register:
            offset = stmt.dst.reg_offset
            data = self._expr(stmt.src)
            size = stmt.src.bits // 8

            self._assign_to_register(offset, data, size, src=stmt.src, dst=stmt.dst)

        elif dst_type is ailment.Expr.Tmp:
            # simply write to self.tmps
            data = self._expr(stmt.src)
            if data is None:
                return

            self.tmps[stmt.dst.tmp_idx] = data

        else:
            l.warning('Unsupported dst type %s.', dst_type)

    def _ail_handle_Store(self, stmt: ailment.Stmt.Store):
        addr_r = self._expr(stmt.addr)
        data = self._expr(stmt.data)
        size = stmt.size
        self._store(addr_r, data, size, stmt=stmt)

    def _ail_handle_Jump(self, stmt):
        pass

    def _ail_handle_ConditionalJump(self, stmt):
        self._expr(stmt.condition)

    def _ail_handle_Call(self, stmt: ailment.Stmt.Call, is_expr=False) -> Optional[RichR]:
        target = stmt.target
        args = [ ]
        if stmt.args:
            for arg in stmt.args:
                self._reference_spoffset = True
                richr = self._expr(arg)
                self._reference_spoffset = False
                args.append(richr)

        ret_expr = None
        ret_reg_offset = None
        ret_val = None  # stores the value that this method should return to its caller when this is a call expression.
        if not is_expr:
            # this is a call statement. we need to update the return value register later
            ret_expr: Optional[ailment.Expr.Register] = stmt.ret_expr
            if ret_expr is not None:
                ret_reg_offset = ret_expr.reg_offset
            else:
                if stmt.calling_convention is not None:
                    if stmt.prototype is None:
                        ret_expr: SimRegArg = stmt.calling_convention.RETURN_VAL
                    elif stmt.prototype.returnty is None or type(stmt.prototype.returnty) is SimTypeBottom:
                        ret_expr = None
                    else:
                        ret_expr: SimRegArg = stmt.calling_convention.return_val(stmt.prototype.returnty)
                else:
                    l.debug("Unknown calling convention for function %s. Fall back to default calling convention.",
                            target)
                    ret_expr: SimRegArg = self.project.factory.cc().RETURN_VAL

                if ret_expr is not None:
                    ret_reg_offset = self.project.arch.registers[ret_expr.reg_name][0]
        else:
            # this is a call expression. we just return the value at the end of this method
            pass

        # discover the prototype
        prototype: Optional[SimTypeFunction] = None
        if stmt.prototype is not None:
            prototype = stmt.prototype
        elif isinstance(stmt.target, ailment.Expr.Const):
            func_addr = stmt.target.value
            if func_addr in self.kb.functions:
                func = self.kb.functions[func_addr]
                prototype = func.prototype

        # dump the type of the return value
        if prototype is not None:
            ret_ty = TypeLifter(self.arch.bits).lift(prototype.returnty)
        else:
            ret_ty = typevars.TypeVariable()
        if isinstance(ret_ty, typeconsts.BottomType):
            ret_ty = typevars.TypeVariable()

        if is_expr:
            # call expression mode
            ret_val = RichR(self.state.top(self.state.arch.bits), typevar=ret_ty)
        else:
            if ret_expr is not None:
                # update the return value register
                self._assign_to_register(
                    ret_reg_offset,
                    RichR(self.state.top(self.state.arch.bits), typevar=ret_ty),
                    self.state.arch.bytes,
                    dst=ret_expr,
                )

        if prototype is not None and args:
            # add type constraints
            for arg, arg_type in zip(args, prototype.args):
                if arg.typevar is not None:
                    arg_ty = TypeLifter(self.arch.bits).lift(arg_type)
                    type_constraint = typevars.Subtype(
                        arg.typevar, arg_ty
                    )
                    self.state.add_type_constraint(type_constraint)

        if is_expr:
            # call expression mode: return the actual return value
            return ret_val
        return None

    def _ail_handle_CallExpr(self, expr: ailment.Stmt.Call) -> RichR:
        return self._ail_handle_Call(expr, is_expr=True)

    def _ail_handle_Return(self, stmt: ailment.Stmt.Return):
        if stmt.ret_exprs:
            for ret_expr in stmt.ret_exprs:
                self._expr(ret_expr)

    # Expression handlers

    def _expr(self, expr: ailment.Expr.Expression):
        """

        :param expr:
        :return:
        :rtype: RichR
        """

        r = super()._expr(expr)
        if r is None:
            return RichR(self.state.top(expr.bits))
        return r

    def _ail_handle_BV(self, expr: claripy.ast.Base):
        return RichR(expr)

    def _ail_handle_Register(self, expr):
        offset = expr.reg_offset
        size = expr.bits // 8

        return self._read_from_register(offset, size, expr=expr)

    def _ail_handle_Load(self, expr):
        addr_r = self._expr(expr.addr)
        size = expr.size

        return self._load(addr_r, size, expr=expr)

    def _ail_handle_Const(self, expr):
        v = claripy.BVV(expr.value, expr.size * self.state.arch.byte_width)
        r = RichR(v, typevar=typeconsts.int_type(expr.size * self.state.arch.byte_width))
        self._reference(r, self._codeloc())
        return r

    def _ail_handle_BinaryOp(self, expr):
        r = super()._ail_handle_BinaryOp(expr)
        if r is None:
            # Treat it as a normal binaryop expression
            self._expr(expr.operands[0])
            self._expr(expr.operands[1])
            # still return a RichR instance
            r = RichR(self.state.top(expr.bits))
        return r

    def _ail_handle_Convert(self, expr: ailment.Expr.Convert):
        r = self._expr(expr.operand)
        typevar = None
        if r.typevar is not None:
            if isinstance(r.typevar, typevars.DerivedTypeVariable) and isinstance(r.typevar.label, typevars.ConvertTo):
                # there is already a conversion - overwrite it
                typevar = typevars.DerivedTypeVariable(r.typevar.type_var, typevars.ConvertTo(expr.to_bits))
            else:
                typevar = typevars.DerivedTypeVariable(r.typevar, typevars.ConvertTo(expr.to_bits))

        return RichR(self.state.top(expr.to_bits), typevar=typevar)

    def _ail_handle_Reinterpret(self, expr: ailment.Expr.Reinterpret):
        r = self._expr(expr.operand)
        typevar = None
        if r.typevar is not None:
            if isinstance(r.typevar, typevars.DerivedTypeVariable) and \
                    isinstance(r.typevar.label, typevars.ReinterpretAs):
                # there is already a reinterpretas - overwrite it
                typevar = typevars.DerivedTypeVariable(
                    r.typevar.type_var,
                    typevars.ReinterpretAs(expr.to_type, expr.to_bits)
                )
            else:
                typevar = typevars.DerivedTypeVariable(r.typevar, typevars.ReinterpretAs(expr.to_type, expr.to_bits))

        return RichR(self.state.top(expr.to_bits), typevar=typevar)

    def _ail_handle_StackBaseOffset(self, expr: ailment.Expr.StackBaseOffset):
        typevar = self.state.stack_offset_typevars.get(expr.offset, None)

        if typevar is None:
            # allocate a new type variable
            typevar = typevars.TypeVariable()
            self.state.stack_offset_typevars[expr.offset] = typevar

        value_v = self.state.stack_address(expr.offset)
        richr = RichR(value_v, typevar=typevar)
        if self._reference_spoffset:
            self._reference(richr, self._codeloc(), src=expr)

        return richr

    def _ail_handle_ITE(self, expr: ailment.Expr.ITE):
        # pylint:disable=unused-variable
        cond = self._expr(expr.cond)
        r0 = self._expr(expr.iftrue)
        r1 = self._expr(expr.iffalse)

        return RichR(self.state.top(expr.bits))

    def _ail_handle_Cmp(self, expr):  # pylint:disable=useless-return
        self._expr(expr.operands[0])
        self._expr(expr.operands[1])
        return RichR(self.state.top(1))

    _ail_handle_CmpF = _ail_handle_Cmp
    _ail_handle_CmpEQ = _ail_handle_Cmp
    _ail_handle_CmpNE = _ail_handle_Cmp
    _ail_handle_CmpLT = _ail_handle_Cmp
    _ail_handle_CmpLE = _ail_handle_Cmp
    _ail_handle_CmpGT = _ail_handle_Cmp
    _ail_handle_CmpGE = _ail_handle_Cmp

    def _ail_handle_Add(self, expr):

        arg0, arg1 = expr.operands

        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        type_constraints = set()
        if r0.typevar is not None and r1.data.concrete:
            # addition with constants. create a derived type variable
            typevar = typevars.DerivedTypeVariable(r0.typevar, typevars.AddN(r1.data._model_concrete.value))
        else:
            # create a new type variable and add constraints accordingly
            typevar = typevars.TypeVariable()
            if r0.typevar is not None and r1.typevar is not None:
                type_constraints.add(typevars.Add(r0.typevar, r1.typevar, typevar))

        sum_ = None
        if r0.data is not None and r1.data is not None:
            sum_ = r0.data + r1.data

        return RichR(sum_,
                     typevar=typevar,
                     type_constraints=type_constraints,
                     )

    def _ail_handle_Sub(self, expr):

        arg0, arg1 = expr.operands

        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        type_constraints = set()
        if r0.typevar is not None and r1.data.concrete:
            typevar = typevars.DerivedTypeVariable(r0.typevar, typevars.SubN(r1.data._model_concrete.value))
        else:
            typevar = typevars.TypeVariable()
            if r0.typevar is not None and r1.typevar is not None:
                type_constraints.add(typevars.Sub(r0.typevar, r1.typevar, typevar))

        sub = None
        if r0.data is not None and r1.data is not None:
            sub = r0.data - r1.data

        return RichR(sub,
                     typevar=typevar,
                     type_constraints=type_constraints,
                     )

    def _ail_handle_Mul(self, expr):

        arg0, arg1 = expr.operands

        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        if r0.data.concrete and r1.data.concrete:
            # constants
            result_size = arg0.bits
            return RichR(r0.data * r1.data,
                         typevar=typeconsts.int_type(result_size),
                         type_constraints=None)

        r = self.state.top(expr.bits)
        return RichR(r,
                     typevar=r0.typevar,
                     )

    def _ail_handle_Mull(self, expr):

        arg0, arg1 = expr.operands

        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        if r0.data.concrete and r1.data.concrete:
            # constants
            result_size = expr.bits
            if r0.data.size() < result_size:
                if expr.signed:
                    r0.data = claripy.SignExt(result_size - r0.data.size(), r0.data)
                else:
                    r0.data = claripy.ZeroExt(result_size - r0.data.size(), r0.data)
            if r1.data.size() < result_size:
                if expr.signed:
                    r1.data = claripy.SignExt(result_size - r1.data.size(), r1.data)
                else:
                    r1.data = claripy.ZeroExt(result_size - r1.data.size(), r1.data)
            return RichR(r0.data * r1.data,
                         typevar=typeconsts.int_type(result_size),
                         type_constraints=None)

        r = self.state.top(expr.bits)
        return RichR(r,
                     typevar=r0.typevar,  # FIXME: the size is probably changed
                     )

    def _ail_handle_Div(self, expr):

        arg0, arg1 = expr.operands

        r0 = self._expr(arg0)
        r1 = self._expr(arg1)
        from_size = expr.bits
        to_size = r1.bits

        if expr.signed:
            remainder = r0.data.SMod(claripy.SignExt(from_size - to_size, r1.data))
        else:
            remainder = r0.data % claripy.ZeroExt(from_size - to_size, r1.data)

        return RichR(remainder,
                     # typevar=r0.typevar,  # FIXME: Handle typevars for Div
                     )

    def _ail_handle_DivMod(self, expr: ailment.Expr.BinaryOp):

        arg0, arg1 = expr.operands

        r0 = self._expr(arg0)
        r1 = self._expr(arg1)
        from_size = expr.bits
        to_size = r1.bits

        if expr.signed:
            quotient = r0.data.SDiv(claripy.SignExt(from_size - to_size, r1.data))
            remainder = r0.data.SMod(claripy.SignExt(from_size - to_size, r1.data))
            quotient_size = to_size
            remainder_size = to_size
            r = claripy.Concat(
                claripy.Extract(remainder_size - 1, 0, remainder),
                claripy.Extract(quotient_size - 1, 0, quotient)
            )
        else:
            quotient = (r0.data // claripy.ZeroExt(from_size - to_size, r1.data))
            remainder = (r0.data % claripy.ZeroExt(from_size - to_size, r1.data))
            quotient_size = to_size
            remainder_size = to_size
            r = claripy.Concat(
                claripy.Extract(remainder_size - 1, 0, remainder),
                claripy.Extract(quotient_size - 1, 0, quotient)
            )

        return RichR(r,
                     # typevar=r0.typevar,  # FIXME: Handle typevars for DivMod
                     )

    def _ail_handle_Xor(self, expr):

        arg0, arg1 = expr.operands

        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        if r0.data.concrete and r1.data.concrete:
            # constants
            result_size = arg0.bits
            return RichR(r0.data ^ r1.data,
                         typevar=typeconsts.int_type(result_size),
                         type_constraints=None)

        r = self.state.top(expr.bits)
        return RichR(r,
                     typevar=r0.typevar,
                     )

    def _ail_handle_Shl(self, expr):

        arg0, arg1 = expr.operands

        r0 = self._expr(arg0)
        r1 = self._expr(arg1)
        result_size = arg0.bits

        if not r1.data.concrete:
            # we don't support symbolic shiftamount
            r = self.state.top(result_size)
            return RichR(r,
                         typevar=r0.typevar,
                         )

        shiftamount = r1.data._model_concrete.value

        return RichR(r0.data << shiftamount,
                     typevar=typeconsts.int_type(result_size),
                     type_constraints=None)

    def _ail_handle_Shr(self, expr):

        arg0, arg1 = expr.operands

        r0 = self._expr(arg0)
        r1 = self._expr(arg1)
        result_size = arg0.bits

        if not r1.data.concrete:
            # we don't support symbolic shiftamount
            r = self.state.top(result_size)
            return RichR(r,
                         typevar=r0.typevar,
                         )

        shiftamount = r1.data._model_concrete.value

        return RichR(claripy.LShR(r0.data, shiftamount),
                     typevar=typeconsts.int_type(result_size),
                     type_constraints=None)

    def _ail_handle_Sal(self, expr):

        arg0, arg1 = expr.operands

        r0 = self._expr(arg0)
        r1 = self._expr(arg1)
        result_size = arg0.bits

        if not r1.data.concrete:
            # we don't support symbolic shiftamount
            r = self.state.top(result_size)
            return RichR(r,
                         typevar=r0.typevar,
                         )

        shiftamount = r1.data._model_concrete.value

        return RichR(r0.data << shiftamount,
                     typevar=typeconsts.int_type(result_size),
                     type_constraints=None)

    def _ail_handle_Sar(self, expr):

        arg0, arg1 = expr.operands

        r0 = self._expr(arg0)
        r1 = self._expr(arg1)
        result_size = arg0.bits

        if not r1.data.concrete:
            # we don't support symbolic shiftamount
            r = self.state.top(result_size)
            return RichR(r,
                         typevar=r0.typevar,
                         )

        shiftamount = r1.data._model_concrete.value

        return RichR(r0.data >> shiftamount,
                     typevar=typeconsts.int_type(result_size),
                     type_constraints=None)

    def _ail_handle_And(self, expr):

        arg0, arg1 = expr.operands

        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        if r0.data.concrete and r1.data.concrete:
            result_size = arg0.bits
            return RichR(
                r0.data & r1.data,
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(expr.bits)
        return RichR(r, typevar=r0.typevar)

    def _ail_handle_Or(self, expr):

        arg0, arg1 = expr.operands

        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        if r0.data.concrete and r1.data.concrete:
            result_size = arg0.bits
            return RichR(
                r0.data | r1.data,
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(expr.bits)
        return RichR(r, typevar=r0.typevar)

    def _ail_handle_Concat(self, expr):

        arg0, arg1 = expr.operands

        _ = self._expr(arg0)
        _ = self._expr(arg1)

        # TODO: Model the operation. Don't lose type constraints
        return RichR(self.state.top(expr.bits))

    def _ail_handle_Not(self, expr):
        arg = expr.operands[0]
        expr = self._expr(arg)


        result_size = arg.bits

        if expr.data.concrete:
            return RichR(
                ~expr.data,
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(result_size)
        return RichR(r, typevar=expr.typevar)



###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_PIC_17_LE_16_PIC_17C7xx(ArchPcode):
    name = 'PIC-17:LE:16:PIC-17C7xx'
    pcode_arch = 'PIC-17:LE:16:PIC-17C7xx'
    description = 'PIC-17C7xx'
    bits = 16
    ip_offset = 0x0
    sp_offset = 0x4
    bp_offset = sp_offset
    instruction_endness = Endness.LE
    register_list = [
        Register('indf0', 1, 0x0),
        Register('fsr0', 1, 0x1),
        Register('pclat', 2, 0x2),
        Register('pcl', 1, 0x2),
        Register('pclath', 1, 0x3),
        Register('alusta', 1, 0x4),
        Register('t0sta', 1, 0x5),
        Register('cpusta', 1, 0x6),
        Register('intsta', 1, 0x7),
        Register('indf1', 1, 0x8),
        Register('fsr1', 1, 0x9),
        Register('tmr0l', 1, 0xb),
        Register('tmr0h', 1, 0xc),
        Register('tblptr', 2, 0xd),
        Register('tblptrl', 1, 0xd),
        Register('tblptrh', 1, 0xe),
        Register('bsr', 1, 0xf),
        Register('porta', 1, 0x10),
        Register('ddrb', 1, 0x11),
        Register('portb', 1, 0x12),
        Register('rcsta1', 1, 0x13),
        Register('rcreg1', 1, 0x14),
        Register('txsta1', 1, 0x15),
        Register('txreg1', 1, 0x16),
        Register('spbrg1', 1, 0x17),
        Register('prod', 2, 0x18),
        Register('prodl', 1, 0x18),
        Register('prodh', 1, 0x19),
        Register('ddrc', 1, 0x110),
        Register('portc', 1, 0x111),
        Register('ddrd', 1, 0x112),
        Register('portd', 1, 0x113),
        Register('ddre', 1, 0x114),
        Register('porte', 1, 0x115),
        Register('pir1', 1, 0x116),
        Register('pie1', 1, 0x117),
        Register('tmr1', 1, 0x210),
        Register('tmr2', 1, 0x211),
        Register('tmr3l', 1, 0x212),
        Register('tmr3h', 1, 0x213),
        Register('pr1', 1, 0x214),
        Register('pr2', 1, 0x215),
        Register('pr3lca1l', 1, 0x216),
        Register('pr3hca1h', 1, 0x217),
        Register('pw1dcl', 1, 0x310),
        Register('pw2dcl', 1, 0x311),
        Register('pw1dch', 1, 0x312),
        Register('pw2dch', 1, 0x313),
        Register('ca2l', 1, 0x314),
        Register('ca2h', 1, 0x315),
        Register('tcon1', 1, 0x316),
        Register('tcon2', 1, 0x317),
        Register('pir2', 1, 0x410),
        Register('pie2', 1, 0x411),
        Register('rcsta2', 1, 0x413),
        Register('rcreg2', 1, 0x414),
        Register('txsta2', 1, 0x415),
        Register('txreg2', 1, 0x416),
        Register('spbrg2', 1, 0x417),
        Register('ddrf', 1, 0x510),
        Register('portf', 1, 0x511),
        Register('ddrg', 1, 0x512),
        Register('portg', 1, 0x513),
        Register('adcon0', 1, 0x514),
        Register('adcon1', 1, 0x515),
        Register('adres', 2, 0x516),
        Register('adresl', 1, 0x516),
        Register('adresh', 1, 0x517),
        Register('sspadd', 1, 0x610),
        Register('sspcon1', 1, 0x611),
        Register('sspcon2', 1, 0x612),
        Register('sspstat', 1, 0x613),
        Register('sspbuf', 1, 0x614),
        Register('pw3dcl', 1, 0x710),
        Register('pw3dch', 1, 0x711),
        Register('ca3l', 1, 0x712),
        Register('ca3h', 1, 0x713),
        Register('ca4l', 1, 0x714),
        Register('ca4h', 1, 0x715),
        Register('tcon3', 1, 0x716),
        Register('ddrh', 1, 0x810),
        Register('porth', 1, 0x811),
        Register('ddrj', 1, 0x812),
        Register('portj', 1, 0x813),
        Register('pc', 2, 0x0, alias_names=('ip',)),
        Register('stkptr', 1, 0x4),
        Register('fs32', 1, 0x5),
        Register('fs10', 1, 0x6),
        Register('ov', 1, 0x7),
        Register('z', 1, 0x8),
        Register('dc', 1, 0x9),
        Register('c', 1, 0xa),
        Register('tblat', 2, 0x10),
        Register('tblatl', 1, 0x10),
        Register('tblath', 1, 0x11),
        Register('wreg', 1, 0x20)
    ]

register_arch(['pic-17:le:16:pic-17c7xx'], 16, Endness.LE, ArchPcode_PIC_17_LE_16_PIC_17C7xx)



import angr

class strcpy(angr.SimProcedure):
    #pylint:disable=arguments-differ

    def run(self, dst, src):
        strlen = angr.SIM_PROCEDURES['libc']['strlen']
        strncpy = angr.SIM_PROCEDURES['libc']['strncpy']
        src_len = self.inline_call(strlen, src)

        ret_expr = self.inline_call(strncpy, dst, src, src_len.ret_expr+1, src_len=src_len.ret_expr).ret_expr
        return ret_expr



import angr
import archinfo
import logging

l = logging.getLogger(name=__name__)

class sigaction(angr.SimProcedure):
    def run(self, signum, act, oldact):
        l.warning('Calling sigaction for signal %s - this is emulated as a nop', signum)
        if not self.state.solver.is_true(oldact == 0):
            if isinstance(self.arch, archinfo.ArchAMD64):
                self.state.memory.store(oldact, self.state.solver.BVS('sigaction_oldact', 152*8, key=('api', 'sigaction', 'oldact')))
            elif isinstance(self.arch, archinfo.ArchX86):
                self.state.memory.store(oldact, self.state.solver.BVS('sigaction_oldact', 140*8, key=('api', 'sigaction', 'oldact')))
        return 0



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("diagnosticdataquery.dll")
prototypes = \
    {
        # 
        'DdqCreateSession': SimTypeFunction([SimTypeInt(signed=False, label="DdqAccessLevel"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["accessLevel", "hSession"]),
        # 
        'DdqCloseSession': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession"]),
        # 
        'DdqGetSessionAccessLevel': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="DdqAccessLevel"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "accessLevel"]),
        # 
        'DdqGetDiagnosticDataAccessLevelAllowed': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="DdqAccessLevel"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["accessLevel"]),
        # 
        'DdqGetDiagnosticRecordStats': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"producerNames": SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0), "producerNameCount": SimTypeInt(signed=False, label="UInt32"), "textToMatch": SimTypePointer(SimTypeChar(label="Char"), offset=0), "categoryIds": SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), "categoryIdCount": SimTypeInt(signed=False, label="UInt32"), "privacyTags": SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), "privacyTagCount": SimTypeInt(signed=False, label="UInt32"), "coreDataOnly": SimTypeInt(signed=True, label="Int32")}, name="DIAGNOSTIC_DATA_SEARCH_CRITERIA", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeLongLong(signed=True, label="Int64"), offset=0), SimTypePointer(SimTypeLongLong(signed=True, label="Int64"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "searchCriteria", "recordCount", "minRowId", "maxRowId"]),
        # 
        'DdqGetDiagnosticRecordPayload': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeLongLong(signed=True, label="Int64"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "rowId", "payload"]),
        # 
        'DdqGetDiagnosticRecordLocaleTags': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "locale", "hTagDescription"]),
        # 
        'DdqFreeDiagnosticRecordLocaleTags': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTagDescription"]),
        # 
        'DdqGetDiagnosticRecordLocaleTagAtIndex': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"privacyTag": SimTypeInt(signed=True, label="Int32"), "name": SimTypePointer(SimTypeChar(label="Char"), offset=0), "description": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="DIAGNOSTIC_DATA_EVENT_TAG_DESCRIPTION", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTagDescription", "index", "tagDescription"]),
        # 
        'DdqGetDiagnosticRecordLocaleTagCount': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTagDescription", "tagDescriptionCount"]),
        # 
        'DdqGetDiagnosticRecordProducers': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "hProducerDescription"]),
        # 
        'DdqFreeDiagnosticRecordProducers': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProducerDescription"]),
        # 
        'DdqGetDiagnosticRecordProducerAtIndex': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"name": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="DIAGNOSTIC_DATA_EVENT_PRODUCER_DESCRIPTION", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProducerDescription", "index", "producerDescription"]),
        # 
        'DdqGetDiagnosticRecordProducerCount': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProducerDescription", "producerDescriptionCount"]),
        # 
        'DdqGetDiagnosticRecordProducerCategories': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "producerName", "hCategoryDescription"]),
        # 
        'DdqFreeDiagnosticRecordProducerCategories': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hCategoryDescription"]),
        # 
        'DdqGetDiagnosticRecordCategoryAtIndex': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"id": SimTypeInt(signed=True, label="Int32"), "name": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="DIAGNOSTIC_DATA_EVENT_CATEGORY_DESCRIPTION", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hCategoryDescription", "index", "categoryDescription"]),
        # 
        'DdqGetDiagnosticRecordCategoryCount': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hCategoryDescription", "categoryDescriptionCount"]),
        # 
        'DdqIsDiagnosticRecordSampledIn': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeLongLong(signed=False, label="UInt64"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "providerGroup", "providerId", "providerName", "eventId", "eventName", "eventVersion", "eventKeywords", "isSampledIn"]),
        # 
        'DdqGetDiagnosticRecordPage': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"producerNames": SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0), "producerNameCount": SimTypeInt(signed=False, label="UInt32"), "textToMatch": SimTypePointer(SimTypeChar(label="Char"), offset=0), "categoryIds": SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), "categoryIdCount": SimTypeInt(signed=False, label="UInt32"), "privacyTags": SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), "privacyTagCount": SimTypeInt(signed=False, label="UInt32"), "coreDataOnly": SimTypeInt(signed=True, label="Int32")}, name="DIAGNOSTIC_DATA_SEARCH_CRITERIA", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeLongLong(signed=True, label="Int64"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "searchCriteria", "offset", "pageRecordCount", "baseRowId", "hRecord"]),
        # 
        'DdqFreeDiagnosticRecordPage': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hRecord"]),
        # 
        'DdqGetDiagnosticRecordAtIndex': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"rowId": SimTypeLongLong(signed=True, label="Int64"), "timestamp": SimTypeLongLong(signed=False, label="UInt64"), "eventKeywords": SimTypeLongLong(signed=False, label="UInt64"), "fullEventName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "providerGroupGuid": SimTypePointer(SimTypeChar(label="Char"), offset=0), "producerName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "privacyTags": SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), "privacyTagCount": SimTypeInt(signed=False, label="UInt32"), "categoryIds": SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), "categoryIdCount": SimTypeInt(signed=False, label="UInt32"), "isCoreData": SimTypeInt(signed=True, label="Int32"), "extra1": SimTypePointer(SimTypeChar(label="Char"), offset=0), "extra2": SimTypePointer(SimTypeChar(label="Char"), offset=0), "extra3": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="DIAGNOSTIC_DATA_RECORD", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hRecord", "index", "record"]),
        # 
        'DdqGetDiagnosticRecordCount': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hRecord", "recordCount"]),
        # 
        'DdqGetDiagnosticReportStoreReportCount': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "reportStoreType", "reportCount"]),
        # 
        'DdqCancelDiagnosticRecordOperation': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession"]),
        # 
        'DdqGetDiagnosticReport': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "reportStoreType", "hReport"]),
        # 
        'DdqFreeDiagnosticReport': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hReport"]),
        # 
        'DdqGetDiagnosticReportAtIndex': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"signature": SimStruct({"eventName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 65), "parameters": SimTypeFixedSizeArray(SimStruct({"name": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 129), "value": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 260)}, name="DIAGNOSTIC_REPORT_PARAMETER", pack=False, align=None), 10)}, name="DIAGNOSTIC_REPORT_SIGNATURE", pack=False, align=None), "bucketId": SimTypeBottom(label="Guid"), "reportId": SimTypeBottom(label="Guid"), "creationTime": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "sizeInBytes": SimTypeLongLong(signed=False, label="UInt64"), "cabId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "reportStatus": SimTypeInt(signed=False, label="UInt32"), "reportIntegratorId": SimTypeBottom(label="Guid"), "fileNames": SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0), "fileCount": SimTypeInt(signed=False, label="UInt32"), "friendlyEventName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "applicationName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "applicationPath": SimTypePointer(SimTypeChar(label="Char"), offset=0), "description": SimTypePointer(SimTypeChar(label="Char"), offset=0), "bucketIdString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "legacyBucketId": SimTypeLongLong(signed=False, label="UInt64"), "reportKey": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="DIAGNOSTIC_REPORT_DATA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hReport", "index", "report"]),
        # 
        'DdqGetDiagnosticReportCount': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hReport", "reportCount"]),
        # 
        'DdqExtractDiagnosticReport': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "reportStoreType", "reportKey", "destinationPath"]),
        # 
        'DdqGetDiagnosticRecordTagDistribution': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"privacyTag": SimTypeInt(signed=True, label="Int32"), "eventCount": SimTypeInt(signed=False, label="UInt32")}, name="DIAGNOSTIC_DATA_EVENT_TAG_STATS", pack=False, align=None), offset=0), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "producerNames", "producerNameCount", "tagStats", "statCount"]),
        # 
        'DdqGetDiagnosticRecordBinaryDistribution': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"moduleName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "friendlyModuleName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "eventCount": SimTypeInt(signed=False, label="UInt32"), "uploadSizeBytes": SimTypeLongLong(signed=False, label="UInt64")}, name="DIAGNOSTIC_DATA_EVENT_BINARY_STATS", pack=False, align=None), offset=0), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "producerNames", "producerNameCount", "topNBinaries", "binaryStats", "statCount"]),
        # 
        'DdqGetDiagnosticRecordSummary': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"optInLevel": SimTypeInt(signed=False, label="UInt32"), "transcriptSizeBytes": SimTypeLongLong(signed=False, label="UInt64"), "oldestEventTimestamp": SimTypeLongLong(signed=False, label="UInt64"), "totalEventCountLast24Hours": SimTypeInt(signed=False, label="UInt32"), "averageDailyEvents": SimTypeFloat(size=32)}, name="DIAGNOSTIC_DATA_GENERAL_STATS", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "producerNames", "producerNameCount", "generalStats"]),
        # 
        'DdqSetTranscriptConfiguration': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"hoursOfHistoryToKeep": SimTypeInt(signed=False, label="UInt32"), "maxStoreMegabytes": SimTypeInt(signed=False, label="UInt32"), "requestedMaxStoreMegabytes": SimTypeInt(signed=False, label="UInt32")}, name="DIAGNOSTIC_DATA_EVENT_TRANSCRIPT_CONFIGURATION", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "desiredConfig"]),
        # 
        'DdqGetTranscriptConfiguration': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"hoursOfHistoryToKeep": SimTypeInt(signed=False, label="UInt32"), "maxStoreMegabytes": SimTypeInt(signed=False, label="UInt32"), "requestedMaxStoreMegabytes": SimTypeInt(signed=False, label="UInt32")}, name="DIAGNOSTIC_DATA_EVENT_TRANSCRIPT_CONFIGURATION", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hSession", "currentConfig"]),
    }

lib.set_prototypes(prototypes)



from .graph import GraphVisitor


class SingleNodeGraphVisitor(GraphVisitor):
    """
    :param node: The single node that should be in the graph.
    """

    __slots__ = ('node', 'node_returned', )

    def __init__(self, node):
        super(SingleNodeGraphVisitor, self).__init__()
        self.node = node
        self.node_returned = False

    def reset(self):
        self.node_returned = False

    def next_node(self):
        if self.node_returned:
            return None
        self.node_returned = True
        return self.node

    def successors(self, node):
        return [ ]

    def predecessors(self, node):
        return [ ]

    def sort_nodes(self, nodes=None):
        if nodes:
            return nodes
        else:
            return [ self.node ]



###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_DATA_BE_64_default(ArchPcode):
    name = 'DATA:BE:64:default'
    pcode_arch = 'DATA:BE:64:default'
    description = 'Raw Data File (Little Endian)'
    bits = 64
    ip_offset = 0x80000000
    sp_offset = 0x0
    bp_offset = sp_offset
    instruction_endness = Endness.BE
    register_list = [
        Register('sp', 8, 0x0),
        Register('r0', 8, 0x8),
        Register('contextreg', 4, 0x100)
    ]

register_arch(['data:be:64:default'], 64, Endness.BE, ArchPcode_DATA_BE_64_default)



# pylint:disable=unused-argument,arguments-differ
from collections import defaultdict
from typing import Dict, List

import ailment

from ..sequence_walker import SequenceWalker
from ..structurer_nodes import SequenceNode, CodeNode, MultiNode, LoopNode, ConditionNode, ContinueNode, \
    CascadingConditionNode


class LoopSimplifier(SequenceWalker):
    """
    Simplifies loops.
    """
    def __init__(self, node):
        handlers = {
            SequenceNode: self._handle_sequencenode,
            CodeNode: self._handle_codenode,
            MultiNode: self._handle_multinode,
            LoopNode: self._handle_loopnode,
            ConditionNode: self._handle_conditionnode,
            CascadingConditionNode: self._handle_cascadingconditionnode,
            ailment.Block: self._handle_block,
        }

        super().__init__(handlers)
        self.continue_preludes: Dict[LoopNode, List[ailment.Block]] = defaultdict(list)
        self.walk(node)

    @staticmethod
    def _control_transferring_statement(stmt: ailment.Stmt.Statement) -> bool:
        return isinstance(stmt,
                          (ailment.Stmt.Call, ailment.Stmt.Return, ailment.Stmt.Jump, ailment.Stmt.ConditionalJump))

    def _handle_sequencenode(self, node, predecessor=None, successor=None, loop=None, loop_successor=None, **kwargs):
        for n0, n1, n2 in zip(node.nodes, node.nodes[1:] + [successor], [predecessor] + node.nodes[:-1]):
            self._handle(n0, predecessor=n2, successor=n1, loop=loop, loop_successor=loop_successor)

    def _handle_codenode(self, node, predecessor=None, successor=None, loop=None, loop_successor=None, **kwargs):
        self._handle(node.node, predecessor=predecessor, successor=successor, loop=loop, loop_successor=loop_successor)

    def _handle_conditionnode(self, node, predecessor=None, successor=None, loop=None, loop_successor=None, **kwargs):
        if node.true_node is not None:
            self._handle(node.true_node, predecessor=predecessor, successor=successor, loop=loop,
                         loop_successor=loop_successor)
        if node.false_node is not None:
            self._handle(node.false_node, predecessor=predecessor, successor=successor, loop=loop,
                         loop_successor=loop_successor)

    def _handle_cascadingconditionnode(self, node: CascadingConditionNode, predecessor=None, successor=None, loop=None,
                                       loop_successor=None, **kwargs):

        for _, child_node in node.condition_and_nodes:
            self._handle(child_node, predecessor=predecessor, successor=successor, loop=loop,
                         loop_successor=loop_successor)
        if node.else_node is not None:
            self._handle(node.else_node, predecessor=predecessor, successor=successor, loop=loop,
                         loop_successor=loop_successor)

    def _handle_loopnode(self, node: LoopNode, predecessor=None, successor=None, loop=None, loop_successor=None,
                         **kwargs):
        self._handle(node.sequence_node, predecessor=predecessor, successor=successor, loop=node,
                     loop_successor=successor)

        # find for-loop iterators
        if node.sort == 'while' and self.continue_preludes[node] and \
                (node.condition is not None or len(self.continue_preludes[node]) > 1):
            if all(block.statements for block in self.continue_preludes[node]) and \
                    all(not self._control_transferring_statement(block.statements[-1])
                        for block in self.continue_preludes[node]) and \
                    all(block.statements[-1] == self.continue_preludes[node][0].statements[-1]
                        for block in self.continue_preludes[node]):
                node.sort = 'for'
                node.iterator = self.continue_preludes[node][0].statements[-1]
                for block in self.continue_preludes[node]:
                    block.statements = block.statements[:-1]

        # find for-loop initializers
        if isinstance(predecessor, MultiNode):
            predecessor = predecessor.nodes[-1]
        if node.sort == 'for' and isinstance(predecessor, ailment.Block) and predecessor.statements and \
                isinstance(predecessor.statements[-1], (ailment.Stmt.Assignment, ailment.Stmt.Store)):
            node.initializer = predecessor.statements[-1]
            predecessor.statements = predecessor.statements[:-1]

    def _handle_multinode(self, node, predecessor=None, successor=None, loop=None, loop_successor=None, **kwargs):
        for n0, n1, n2 in zip(node.nodes, node.nodes[1:] + [successor], [predecessor] + node.nodes[:-1]):
            self._handle(n0, predecessor=n2, successor=n1, loop=loop, loop_successor=loop_successor)

    def _handle_block(self, block, predecessor=None, successor=None, loop=None, loop_successor=None, **kwargs):  # pylint:disable=no-self-use
        if isinstance(successor, ContinueNode) or successor is loop_successor:
            self.continue_preludes[loop].append(block)



import angr

import logging
l = logging.getLogger(name=__name__)

class strcmp(angr.SimProcedure):
    #pylint:disable=arguments-differ

    def run(self, a_addr, b_addr, wchar=False, ignore_case=False):
        strlen = angr.SIM_PROCEDURES['libc']['strlen']

        a_strlen = self.inline_call(strlen, a_addr, wchar=wchar)
        b_strlen = self.inline_call(strlen, b_addr, wchar=wchar)
        maxlen = self.state.solver.BVV(max(a_strlen.max_null_index, b_strlen.max_null_index), self.state.arch.bits)

        strncmp = self.inline_call(angr.SIM_PROCEDURES['libc']['strncmp'], a_addr, b_addr, maxlen, a_len=a_strlen, b_len=b_strlen, wchar=wchar, ignore_case=ignore_case)
        return strncmp.ret_expr



from typing import TYPE_CHECKING, List, Set, Optional, Tuple
import logging

from cle import Symbol

l = logging.getLogger(__name__)

if TYPE_CHECKING:
    from angr.code_location import CodeLocation
    from angr.analyses.reaching_definitions.dep_graph import DepGraph
    from angr.analyses.reaching_definitions.rd_state import ReachingDefinitionsState


# pylint: disable=unused-argument, no-self-use
class FunctionHandler:
    """
    An abstract base class for function handlers.

    To work properly, we expect function handlers to:
      - Be related to a <ReachingDefinitionsAnalysis>;
      - Provide a `handle_local_function` method.
    """

    def hook(self, analysis) -> "FunctionHandler":
        """
        A <FunctionHandler> needs information about the context in which it is executed.
        A <ReachingDefinitionsAnalysis> would "hook" into a handler by calling: `<FunctionHandler>.hook(self)`.

        :param angr.analyses.ReachingDefinitionsAnalysis analysis: A RDA using this <FunctionHandler>.

        :return FunctionHandler:
        """
        return self

    def handle_local_function(self,
                              state: 'ReachingDefinitionsState',
                              function_address: int, call_stack: Optional[List],
                              maximum_local_call_depth: int,
                              visited_blocks: Set[int],
                              dep_graph: 'DepGraph',
                              src_ins_addr: Optional[int] = None,
                              codeloc: Optional['CodeLocation'] = None
                              ) -> Tuple[bool, "ReachingDefinitionsState", "Set[int]", "DepGraph"]:
        """
        :param state: The state at the entry of the function, i.e. the function's input state.
        :param function_address: The address of the function to handle.
        :param call_stack:
        :param maximum_local_call_depth:
        :param visited_blocks: A set of the addresses of the previously visited blocks.
        :param dep_graph: A definition-use graph, where nodes represent definitions, and edges represent uses.
        :param codeloc: The code location of the call to the analysed function.
        """
        l.warning('Please implement the local function handler with your own logic.')
        return False, state, visited_blocks, dep_graph

    def handle_unknown_call(self,
                            state: 'ReachingDefinitionsState',
                            src_codeloc: Optional['CodeLocation'] = None
                            ) -> Tuple[bool, 'ReachingDefinitionsState']:
        """
        Called when the RDA encounters a function call to somewhere really weird.
        E.g. the function address was invalid (not even TOP),
        or the address of the function is outside of the main object, but also not a known symbol
        :param state:
        :param src_codeloc:
        :return:
        """
        l.error('Encountered unknown call. Implement the unknown function handler with your own logic.')
        return False, state

    def handle_indirect_call(self,
                             state: 'ReachingDefinitionsState',
                             src_codeloc: Optional['CodeLocation'] = None
                             ) -> Tuple[bool, 'ReachingDefinitionsState']:
        """
        The RDA encountered a function call with multiple possible values, or TOP as a target
        :param state:
        :param src_codeloc:
        :return:
        """
        l.warning('Please implement the indirect function handler with your own logic.')
        return False, state

    def handle_external_function_fallback(self,
                                          state: 'ReachingDefinitionsState',
                                          src_codeloc: Optional['CodeLocation'] = None
                                          ) -> Tuple[bool, 'ReachingDefinitionsState']:
        """
        Fallback for a call to an external function, that has no specific implementation
        :param state:
        :param src_codeloc:
        :return:
        """
        return False, state

    def handle_external_function_symbol(self,
                                 state: 'ReachingDefinitionsState',
                                 symbol: Symbol,
                                 src_codeloc: Optional['CodeLocation'] = None,
                                 ) -> Tuple[bool, 'ReachingDefinitionsState']:
        """
        The generic handler for external functions with a known symbol
        This is different from
        The default behavior using hasattr/getattr supports existing code,
        but you can also implement the check if the external function is supported in another way,
        e.g. similar to SimProcedures
        :param state:
        :param symbol:
        :param src_codeloc:
        :return:
        """
        if symbol.name:
            return self.handle_external_function_name(state, symbol.name, src_codeloc)
        else:
            l.warning('Symbol %s for external function has no name, falling back to generic handler',
                      symbol)
            return self.handle_external_function_fallback(state, src_codeloc)

    def handle_external_function_name(self,
                                      state: 'ReachingDefinitionsState',
                                      ext_func_name: str,
                                      src_codeloc: Optional['CodeLocation'] = None,
                                      ) -> Tuple[bool, 'ReachingDefinitionsState']:
        handler_name = 'handle_%s' % ext_func_name
        if ext_func_name and hasattr(self, handler_name):
            return getattr(self, handler_name)(state, src_codeloc)
        else:
            l.warning('No handler for external function %s(), falling back to generic handler',
                      ext_func_name)
            return self.handle_external_function_fallback(state, src_codeloc)



from ailment.expression import ITE, BinaryOp

from .base import PeepholeOptimizationExprBase


class RemoveRedundantITEBranches(PeepholeOptimizationExprBase):
    __slots__ = ()

    name = "Remove redundant ITE branches"
    expr_classes = (ITE, )  # all expressions are allowed

    def optimize(self, expr: ITE):

        # ITE(cond, a, ITE(!cond, b, c)) ==> ITE(cond, a, b)
        if isinstance(expr.iffalse, ITE):
            # cascading ITEs
            if isinstance(expr.cond, BinaryOp) and isinstance(expr.iffalse.cond, BinaryOp):
                # are they negating each other?
                if expr.cond.op in BinaryOp.COMPARISON_NEGATION and expr.iffalse.cond.op == BinaryOp.COMPARISON_NEGATION[expr.cond.op]:
                    cond0_operands = expr.cond.operands
                    cond1_operands = expr.iffalse.cond.operands
                    if cond0_operands[0].likes(cond1_operands[0]) and cond0_operands[1].likes(cond1_operands[1]):
                        # YES...
                        expr = ITE(expr.idx, expr.cond, expr.iffalse.iftrue, expr.iftrue, **expr.tags)
                        return expr

        return None



# pylint:disable=no-member
import pickle
import logging
from typing import Optional, List, Dict, Tuple
from collections import defaultdict
import bisect

import networkx

from ...misc.ux import once
from ...protos import cfg_pb2, primitives_pb2
from ...serializable import Serializable
from ...utils.enums_conv import cfg_jumpkind_to_pb, cfg_jumpkind_from_pb
from ...errors import AngrCFGError
from .cfg_node import CFGNode
from .memory_data import MemoryData
from .indirect_jump import IndirectJump


l = logging.getLogger(name=__name__)


class CFGModel(Serializable):
    """
    This class describes a Control Flow Graph for a specific range of code.
    """

    __slots__ = ('ident', 'graph', 'jump_tables', 'memory_data', 'insn_addr_to_memory_data', '_nodes_by_addr',
                 '_nodes', '_cfg_manager', '_iropt_level', '_node_addrs', 'is_arm')

    def __init__(self, ident, cfg_manager=None, is_arm=False):

        self.ident = ident
        self._cfg_manager = cfg_manager
        self.is_arm = is_arm

        # Necessary settings
        self._iropt_level = None

        # The graph
        self.graph = networkx.DiGraph()

        # Jump tables
        self.jump_tables: Dict[int,IndirectJump] = { }

        # Memory references
        # A mapping between address and the actual data in memory
        self.memory_data: Dict[int,MemoryData] = { }
        # A mapping between address of the instruction that's referencing the memory data and the memory data itself
        self.insn_addr_to_memory_data: Dict[int,MemoryData] = { }

        # Lists of CFGNodes indexed by the address of each block. Don't serialize
        self._nodes_by_addr = defaultdict(list)
        # CFGNodes dict indexed by block ID. Don't serialize
        self._nodes = { }
        # addresses of CFGNodes to speed up get_any_node(..., anyaddr=True). Don't serialize
        self._node_addrs = [ ]

    #
    # Properties
    #

    @property
    def project(self):
        if self._cfg_manager is None:
            return None
        return self._cfg_manager._kb._project

    #
    # Serialization
    #

    def __getstate__(self):
        state = dict(map(
            lambda x: (x, self.__getattribute__(x)),
            self.__slots__
        ))

        return state

    def __setstate__(self, state):
        for attribute, value in state.items():
            self.__setattr__(attribute, value)

        for addr in self._nodes:
            node = self._nodes[addr]
            node._cfg_model = self

    @classmethod
    def _get_cmsg(cls):
        return cfg_pb2.CFG()

    def serialize_to_cmessage(self):
        if "Emulated" in self.ident:
            raise NotImplementedError("Serializing a CFGEmulated instance is currently not supported.")

        cmsg = self._get_cmsg()
        cmsg.ident = self.ident

        # nodes
        nodes = [ ]
        for n in self.graph.nodes():
            nodes.append(n.serialize_to_cmessage())
        cmsg.nodes.extend(nodes)

        # edges
        edges = [ ]
        for src, dst, data in self.graph.edges(data=True):
            edge = primitives_pb2.Edge()
            edge.src_ea = src.addr
            edge.dst_ea = dst.addr
            for k, v in data.items():
                if k == 'jumpkind':
                    edge.jumpkind = cfg_jumpkind_to_pb(v)
                elif k == 'ins_addr':
                    edge.ins_addr = v if v is not None else 0xffff_ffff_ffff_ffff
                elif k == 'stmt_idx':
                    edge.stmt_idx = v if v is not None else -1
                else:
                    edge.data[k] = pickle.dumps(v)
            edges.append(edge)
        cmsg.edges.extend(edges)

        # memory data
        memory_data = [ ]
        for data in self.memory_data.values():
            memory_data.append(data.serialize_to_cmessage())
        cmsg.memory_data.extend(memory_data)

        return cmsg

    @classmethod
    def parse_from_cmessage(cls, cmsg, cfg_manager=None, loader=None):  # pylint:disable=arguments-differ
        if cfg_manager is None:
            # create a new model unassociated from any project
            model = cls(cmsg.ident)
        else:
            model = cfg_manager.new_model(cmsg.ident)

        # nodes
        for node_pb2 in cmsg.nodes:
            node = CFGNode.parse_from_cmessage(node_pb2, cfg=model)
            model._nodes[node.block_id] = node
            model._nodes_by_addr[node.addr].append(node)
            model.graph.add_node(node)
            if len(model._nodes_by_addr[node.block_id]) > 1:
                if once("cfg_model_parse_from_cmessage many nodes at addr"):
                    l.warning("Importing a CFG with more than one node for a given address is currently unsupported. "
                              "The resulting graph may be broken.")

        model._node_addrs = list(sorted(model._nodes_by_addr.keys()))

        # edges
        for edge_pb2 in cmsg.edges:
            # more than one node at a given address is unsupported, grab the first one
            src = model._nodes_by_addr[edge_pb2.src_ea][0]
            dst = model._nodes_by_addr[edge_pb2.dst_ea][0]
            data = { }
            for k, v in edge_pb2.data.items():
                data[k] = pickle.loads(v)
            data['jumpkind'] = cfg_jumpkind_from_pb(edge_pb2.jumpkind)
            data['ins_addr'] = edge_pb2.ins_addr if edge_pb2.ins_addr != 0xffff_ffff_ffff_ffff else None
            data['stmt_idx'] = edge_pb2.stmt_idx if edge_pb2.stmt_idx != -1 else None
            model.graph.add_edge(src, dst, **data)

        # memory data
        for data_pb2 in cmsg.memory_data:
            md = MemoryData.parse_from_cmessage(data_pb2)
            if loader is not None and md.content is None:
                # fill in the content
                md.fill_content(loader)
            model.memory_data[md.addr] = md

        return model

    #
    # Other methods
    #

    def copy(self):
        model = CFGModel(self.ident, cfg_manager=self._cfg_manager, is_arm=self.is_arm)
        model.graph = networkx.DiGraph(self.graph)
        model.jump_tables = self.jump_tables.copy()
        model.memory_data = self.memory_data.copy()
        model.insn_addr_to_memory_data = self.insn_addr_to_memory_data.copy()
        model._nodes_by_addr = self._nodes_by_addr.copy()
        model._nodes = self._nodes.copy()

        return model

    #
    # Node insertion and removal
    #

    def add_node(self, block_id: int, node: CFGNode) -> None:
        self._nodes[block_id] = node
        self._nodes_by_addr[node.addr].append(node)

        if isinstance(node.addr, int):
            pos = bisect.bisect_left(self._node_addrs, node.addr)
            if pos >= len(self._node_addrs):
                self._node_addrs.append(node.addr)
            elif self._node_addrs[pos] != node.addr:
                self._node_addrs.insert(pos, node.addr)

    def remove_node(self, block_id: int, node: CFGNode) -> None:
        if block_id in self._nodes:
            del self._nodes[block_id]

        if node.addr in self._nodes_by_addr and node in self._nodes_by_addr[node.addr]:
            self._nodes_by_addr[node.addr].remove(node)
            if not self._nodes_by_addr[node.addr]:
                del self._nodes_by_addr[node.addr]

                if isinstance(node.addr, int):
                    pos = bisect.bisect_left(self._node_addrs, node.addr)
                    if pos < len(self._node_addrs) and self._node_addrs[pos] == node.addr:
                        self._node_addrs.pop(pos)

    #
    # CFG View
    #

    def get_node(self, block_id):
        """
        Get a single node from node key.

        :param BlockID block_id: Block ID of the node.
        :return:                 The CFGNode
        :rtype:                  CFGNode
        """
        if block_id in self._nodes:
            return self._nodes[block_id]
        return None

    def get_any_node(self, addr: int, is_syscall: bool=None, anyaddr: bool=False,
                     force_fastpath: bool=False) -> Optional[CFGNode]:
        """
        Get an arbitrary CFGNode (without considering their contexts) from our graph.

        :param addr:            Address of the beginning of the basic block. Set anyaddr to True to support arbitrary
                                address.
        :param is_syscall:      Whether you want to get the syscall node or any other node. This is due to the fact that
                                syscall SimProcedures have the same address as the targer it returns to.
                                None means get either, True means get a syscall node, False means get something that
                                isn't a syscall node.
        :param anyaddr:         If anyaddr is True, then addr doesn't have to be the beginning address of a basic
                                block. By default the entire graph.nodes() will be iterated, and the first node
                                containing the specific address is returned, which can be slow.
        :param force_fastpath:  If force_fastpath is True, it will only perform a dict lookup in the _nodes_by_addr
                                dict.
        :return:                A CFGNode if there is any that satisfies given conditions, or None otherwise
        """

        # fastpath: directly look in the nodes list
        if not anyaddr or addr in self._nodes_by_addr:
            try:
                return self._nodes_by_addr[addr][0]
            except (KeyError, IndexError):
                pass

        if force_fastpath:
            return None

        if isinstance(addr, int):
            # slower path
            # find all potential addresses that the block may cover
            pos = bisect.bisect_left(self._node_addrs, max(addr - 400, 0))

            is_cfgemulated = self.ident == "CFGEmulated"

            while pos < len(self._node_addrs):
                n = self._nodes_by_addr[self._node_addrs[pos]][0]
                actual_addr = n.addr if not self.is_arm else n.addr & 0xffff_fffe
                if actual_addr > addr:
                    break

                if is_cfgemulated:
                    cond = n.looping_times == 0
                else:
                    cond = True
                if anyaddr and n.size is not None:
                    cond = cond and (addr == actual_addr or actual_addr <= addr < actual_addr + n.size)
                else:
                    cond = cond and (addr == actual_addr)
                if cond:
                    if is_syscall is None:
                        return n
                    if n.is_syscall == is_syscall:
                        return n

                pos += 1

        return None

    def get_all_nodes(self, addr: int, is_syscall: bool=None, anyaddr: bool=False) -> List[CFGNode]:
        """
        Get all CFGNodes whose address is the specified one.

        :param addr:       Address of the node
        :param is_syscall: True returns the syscall node, False returns the normal CFGNode, None returns both
        :return:           all CFGNodes
        """
        results = [ ]

        for cfg_node in self.graph.nodes():
            if cfg_node.addr == addr or (anyaddr and
                                         cfg_node.size is not None and
                                         cfg_node.addr <= addr < (cfg_node.addr + cfg_node.size)
                                         ):
                if is_syscall is None or is_syscall == cfg_node.is_syscall:
                    results.append(cfg_node)

        return results

    def nodes(self):
        """
        An iterator of all nodes in the graph.

        :return: The iterator.
        :rtype: iterator
        """

        return self.graph.nodes()

    def get_predecessors(self, cfgnode: CFGNode, excluding_fakeret: bool=True,
                         jumpkind: Optional[str]=None) -> List[CFGNode]:
        """
        Get predecessors of a node in the control flow graph.

        :param cfgnode:             The node.
        :param excluding_fakeret:   True if you want to exclude all predecessors that is connected to the node with a
                                    fakeret edge.
        :param jumpkind:            Only return predecessors with the specified jumpkind. This argument will be ignored
                                    if set to None.
        :return:                    A list of predecessors
        """

        if excluding_fakeret and jumpkind == 'Ijk_FakeRet':
            return [ ]

        if not excluding_fakeret and jumpkind is None:
            # fast path
            if cfgnode in self.graph:
                return list(self.graph.predecessors(cfgnode))
            return [ ]

        predecessors = []
        for pred, _, data in self.graph.in_edges([cfgnode], data=True):
            jk = data['jumpkind']
            if jumpkind is not None:
                if jk == jumpkind:
                    predecessors.append(pred)
            elif excluding_fakeret:
                if jk != 'Ijk_FakeRet':
                    predecessors.append(pred)
            else:
                predecessors.append(pred)
        return predecessors

    def get_successors(self, node: CFGNode,
                       excluding_fakeret: bool = True,
                       jumpkind: Optional[str] = None
                       ) -> List[CFGNode]:
        """
        Get successors of a node in the control flow graph.

        :param CFGNode node:                The node.
        :param bool excluding_fakeret:      True if you want to exclude all successors that is connected to the node
                                            with a fakeret edge.
        :param str or None jumpkind:        Only return successors with the specified jumpkind. This argument will be
                                            ignored if set to None.
        :return:                            A list of successors
        :rtype:                             list
        """

        if jumpkind is not None:
            if excluding_fakeret and jumpkind == 'Ijk_FakeRet':
                return [ ]

        if not excluding_fakeret and jumpkind is None:
            # fast path
            if node in self.graph:
                return list(self.graph.successors(node))
            return [ ]

        successors = []
        for _, suc, data in self.graph.out_edges([node], data=True):
            jk = data['jumpkind']
            if jumpkind is not None:
                if jumpkind == jk:
                    successors.append(suc)
            elif excluding_fakeret:
                if jk != 'Ijk_FakeRet':
                    successors.append(suc)
            else:
                successors.append(suc)
        return successors

    def get_successors_and_jumpkinds(self, node, excluding_fakeret=True):
        """
        Get a list of tuples where the first element is the successor of the CFG node and the second element is the
        jumpkind of the successor.

        :param CFGNode node:            The node.
        :param bool excluding_fakeret:  True if you want to exclude all successors that are fall-through successors.
        :return:                        A list of successors and their corresponding jumpkinds.
        :rtype:                         list
        """

        successors = []
        for _, suc, data in self.graph.out_edges([node], data=True):
            if not excluding_fakeret or data['jumpkind'] != 'Ijk_FakeRet':
                successors.append((suc, data['jumpkind']))
        return successors

    get_successors_and_jumpkind = get_successors_and_jumpkinds

    def get_predecessors_and_jumpkinds(self, node: CFGNode, excluding_fakeret: bool=True) -> List[Tuple[CFGNode,str]]:
        """
        Get a list of tuples where the first element is the predecessor of the CFG node and the second element is the
        jumpkind of the predecessor.

        :param node:                The node.
        :param excluding_fakeret:   True if you want to exclude all predecessors that are fall-through predecessors.
        :return:                    A list of predecessors and their corresponding jumpkinds.
        """

        predecessors = []
        for pred, _, data in self.graph.in_edges([node], data=True):
            if not excluding_fakeret or data['jumpkind'] != 'Ijk_FakeRet':
                predecessors.append((pred, data['jumpkind']))
        return predecessors

    get_predecessors_and_jumpkind = get_predecessors_and_jumpkinds

    def get_all_predecessors(self, cfgnode, depth_limit=None):
        """
        Get all predecessors of a specific node on the control flow graph.

        :param CFGNode cfgnode: The CFGNode object
        :param int depth_limit: Optional depth limit for the depth-first search
        :return: A list of predecessors in the CFG
        :rtype: list
        """
        # use the reverse graph and query for successors (networkx.dfs_predecessors is misleading)
        # dfs_successors returns a dict of (node, [predecessors]). We ignore the keyset and use the values
        predecessors = set().union(*networkx.dfs_successors(self.graph.reverse(), cfgnode, depth_limit).values())
        return list(predecessors)

    def get_all_successors(self, cfgnode, depth_limit=None):
        """
        Get all successors of a specific node on the control flow graph.

        :param CFGNode cfgnode: The CFGNode object
        :param int depth_limit: Optional depth limit for the depth-first search
        :return: A list of successors in the CFG
        :rtype: list
        """
        # dfs_successors returns a dict of (node, [predecessors]). We ignore the keyset and use the values
        successors = set().union(*networkx.dfs_successors(self.graph, cfgnode, depth_limit).values())
        return list(successors)

    def get_branching_nodes(self):
        """
        Returns all nodes that has an out degree >= 2
        """
        nodes = set()
        for n in self.graph.nodes():
            if self.graph.out_degree(n) >= 2:
                nodes.add(n)
        return nodes

    def get_exit_stmt_idx(self, src_block, dst_block):
        """
        Get the corresponding exit statement ID for control flow to reach destination block from source block. The exit
        statement ID was put on the edge when creating the CFG.
        Note that there must be a direct edge between the two blocks, otherwise an exception will be raised.

        :return: The exit statement ID
        """

        if not self.graph.has_edge(src_block, dst_block):
            raise AngrCFGError('Edge (%s, %s) does not exist in CFG' % (src_block, dst_block))

        return self.graph[src_block][dst_block]['stmt_idx']



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("msimg32.dll")
prototypes = \
    {
        # 
        'AlphaBlend': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimStruct({"BlendOp": SimTypeChar(label="Byte"), "BlendFlags": SimTypeChar(label="Byte"), "SourceConstantAlpha": SimTypeChar(label="Byte"), "AlphaFormat": SimTypeChar(label="Byte")}, name="BLENDFUNCTION", pack=False, align=None)], SimTypeInt(signed=True, label="Int32"), arg_names=["hdcDest", "xoriginDest", "yoriginDest", "wDest", "hDest", "hdcSrc", "xoriginSrc", "yoriginSrc", "wSrc", "hSrc", "ftn"]),
        # 
        'TransparentBlt': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hdcDest", "xoriginDest", "yoriginDest", "wDest", "hDest", "hdcSrc", "xoriginSrc", "yoriginSrc", "wSrc", "hSrc", "crTransparent"]),
        # 
        'GradientFill': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"x": SimTypeInt(signed=True, label="Int32"), "y": SimTypeInt(signed=True, label="Int32"), "Red": SimTypeShort(signed=False, label="UInt16"), "Green": SimTypeShort(signed=False, label="UInt16"), "Blue": SimTypeShort(signed=False, label="UInt16"), "Alpha": SimTypeShort(signed=False, label="UInt16")}, name="TRIVERTEX", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="GRADIENT_FILL")], SimTypeInt(signed=True, label="Int32"), arg_names=["hdc", "pVertex", "nVertex", "pMesh", "nMesh", "ulMode"]),
    }

lib.set_prototypes(prototypes)



import angr
import logging

l = logging.getLogger(name=__name__)
#l.setLevel("DEBUG")

""" BOOL GetLastInputInfo(
  PLASTINPUTINFO plii
);

typedef struct tagLASTINPUTINFO {
  UINT  cbSize;  // The size of the structure, in bytes.
                 // This member must be set to sizeof(LASTINPUTINFO).
  DWORD dwTime;
} LASTINPUTINFO, *PLASTINPUTINFO;
"""
class GetLastInputInfo(angr.SimProcedure):
    cbSize = None
    dwTime = None

    def run(self, plii):  #pylint:disable=arguments-differ
        self.fill_symbolic()
        l.info("Setting symbolic memory at %s", str(plii))
        self.state.mem[plii].dword = self.cbSize
        self.state.mem[plii+4].dword = self.dwTime

        return 1

    def fill_symbolic(self):
        self.cbSize = self.state.solver.BVS('tagLASTINPUTINFO_cbSize', 32, key=('api', 'tagLASTINPUTINFO_cbSize'))
        self.dwTime = self.state.solver.BVS('tagLASTINPUTINFO_dwTime', 32, key=('api', 'tagLASTINPUTINFO_dwTime'))

    def fill_concrete(self):
        self.cbSize = self.state.solver.BVV(3, 32)
        self.dwTime = self.state.solver.BVV(3, 32)



import sys
from typing import Sequence, Optional, Callable


if sys.platform == 'win32':
    import colorama  # pylint:disable=import-error


ansi_color_enabled: bool = False


def setup_terminal():
    """
    Check if we are running in a TTY. If so, make sure the terminal supports ANSI escape sequences. If not, disable
    colorized output. Sets global `ansi_color_enabled` to True if colorized output should be enabled by default.
    """
    isatty = (hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
               and hasattr(sys.stderr, 'isatty') and sys.stderr.isatty())
    if sys.platform == 'win32' and isatty:
        if not isinstance(sys.stdout, colorama.ansitowin32.StreamWrapper):
            colorama.init()

    global ansi_color_enabled  # pylint:disable=global-statement
    ansi_color_enabled = isatty


def ansi_color(s: str, color: Optional[str]) -> str:
    """
    Colorize string `s` by wrapping in ANSI escape sequence for given `color`.

    This function does not consider whether escape sequences are functional or not; it is up to the caller to determine
    if its appropriate. Check global `ansi_color_enabled` value in this module.
    """
    if color is None:
        return s

    codes = {
        'black':          '30m',
        'bright_black':   '90m',
        'gray':           '90m',  # alias 'bright black'
        'blue':           '34m',
        'bright_blue':    '94m',
        'cyan':           '36m',
        'bright_cyan':    '96m',
        'green':          '32m',
        'bright_green':   '92m',
        'magenta':        '35m',
        'bright_magenta': '95m',
        'red':            '31m',
        'bright_red':     '91m',
        'white':          '37m',
        'bright_white':   '97m',
        'yellow':         '33m',
        'bright_yellow':  '93m',
    }
    return '\u001b[' + codes[color] + s + '\u001b[0m'


def add_edge_to_buffer(buf: Sequence[str], ref: Sequence[str], start: int, end: int,
                       formatter: Optional[Callable[[str], str]] = None,
                       dashed: bool = False, ascii_only: Optional[bool] = None):
    """
    Draw an edge by adding Unicode box and arrow glyphs to beginning of each line in a list of lines.

    :param buf: Output buffer, used to render formatted edges.
    :param ref: Reference buffer, used to calculate edge depth.
    :param start: Start line.
    :param end: End line, where arrow points.
    :param formatter: Optional callback function used to format the edge before writing it to output buffer.
    :param dashed: Render edge line dashed instead of solid.
    :param ascii_only: Render edge using ASCII characters only. If unspecified, guess by stdout encoding.
    :return:
    """
    abs_start  = min(start, end)
    abs_end    = max(start, end)
    max_depth  = max(map(len, ref[abs_start:abs_end+1]))
    descending = start < end

    if ascii_only is None:
        # Guess whether we should only use ASCII characters based on stdout encoding
        ascii_only = getattr(sys.stdout, 'encoding', None) != 'utf-8'

    if ascii_only:
        chars = {
            'start_cap'    : '-',
            'start_corner' : '+',
            'end_cap'      : '>',
            'end_corner'   : '+',
            'horizontal'   : '+' if dashed else '-',
            'vertical'     : '+' if dashed else '|',
            'spin'         : '@ ',
        }
    else:
        chars = {
            'start_cap'    : '╴',
            'start_corner' : '╭' if descending else '╰',
            'end_cap'      : '⧽',
            'end_corner'   : '╰' if descending else '╭',
            'horizontal'   : '╌' if dashed else '─',
            'vertical'     : '╎' if dashed else '│',
            'spin'         : '⟳ ',
        }

    def handle_line(i, edge_str):
        if formatter is not None:
            edge_str = formatter(edge_str)
        buf[i] = edge_str + buf[i]

    if start == end:
        handle_line(start, chars['spin'])
    else:
        handle_line(start, (chars['start_corner']
                            + chars['horizontal'] * (max_depth - len(ref[start]))
                            + chars['start_cap']))
        handle_line(end, (chars['end_corner']
                          + chars['horizontal'] * (max_depth - len(ref[end]))
                          + chars['end_cap']))
        for i in range(abs_start + 1, abs_end):
            handle_line(i, chars['vertical'] + ' ' * (1 + max_depth - len(ref[i])))



from typing import Optional

from ...state_plugins.sim_action import SimActionData, SimActionObject
from ... import sim_options as o
from . import MemoryMixin


class ActionsMixinHigh(MemoryMixin):
    def load(self, addr, size=None, condition=None, fallback=None, disable_actions=False, action=None, **kwargs):
        if not disable_actions and o.AUTO_REFS in self.state.options and action is None:
            action = self.__make_action('read', addr, size, None, condition, fallback)

        r = super().load(addr, size=size, condition=condition, fallback=fallback, action=action, **kwargs)

        if not disable_actions:
            if o.AST_DEPS in self.state.options and self.category == 'reg':
                r = SimActionObject(r, reg_deps=frozenset((addr,)))

            if action is not None and action.actual_addrs and o.AUTO_REFS in self.state.options:
                # ActionsMixinLow fills up action.actual_addrs. If the load fails, actual_addrs will be None or empty.
                # In that case, we do not add the action.
                action.data = action._make_object(r)
                if action.size is None:
                    action.size = len(r)

                self.state.history.add_action(action)

        return r

    def store(self, addr, data, size=None, disable_actions=False, action=None, condition=None, **kwargs):
        if not disable_actions and o.AUTO_REFS in self.state.options and action is None:
            action = self.__make_action('write', addr, size, data, condition, None)

        super().store(addr, data, size=size, action=action, condition=condition, **kwargs)

        if action is not None and not disable_actions and action.actual_addrs and o.AUTO_REFS in self.state.options:
            # ActionsMixinLow fills up action.actual_addrs. If the store fails, actual_addrs will be None or empty.
            # In that case, we do not add the action.
            self.state.history.add_action(action)

    def __make_action(self, kind, addr, size, data, condition, fallback):
        ref_size = size * self.state.arch.byte_width if size is not None else len(data) if data is not None else None
        region_type = self.category if self.category != "file" else self.id
        action = SimActionData(self.state, region_type, kind,
                               addr=addr,
                               data=data,
                               size=ref_size,
                               condition=condition,
                               fallback=fallback)

        action.added_constraints = self.state.solver.true
        return action

    def _add_constraints(self, c, action=None, **kwargs):
        if action is not None:
            action.added_constraints = self.state.solver.And(action.added_constraints, c)
        return super()._add_constraints(c, action=action, **kwargs)


class ActionsMixinLow(MemoryMixin):
    def load(self, addr, action=None, **kwargs):
        if action is not None:
            if action.actual_addrs is None:
                action.actual_addrs = [ ]
            action.actual_addrs.append(addr)
        return super().load(addr, action=action, **kwargs)

    def store(self, addr, data, action: Optional[SimActionData]=None, **kwargs):
        if action is not None:
            if action.actual_addrs is None:
                action.actual_addrs = [ ]
            action.actual_addrs.append(addr)
            action.actual_value = action._make_object(data)
        return super().store(addr, data, action=action, **kwargs)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("websocket.dll")
prototypes = \
    {
        # 
        'WebSocketCreateClientHandle': SimTypeFunction([SimTypePointer(SimStruct({"Type": SimTypeInt(signed=False, label="WEB_SOCKET_PROPERTY_TYPE"), "pvValue": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "ulValueSize": SimTypeInt(signed=False, label="UInt32")}, name="WEB_SOCKET_PROPERTY", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pProperties", "ulPropertyCount", "phWebSocket"]),
        # 
        'WebSocketBeginClientHandshake': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"pcName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulNameLength": SimTypeInt(signed=False, label="UInt32"), "pcValue": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulValueLength": SimTypeInt(signed=False, label="UInt32")}, name="WEB_SOCKET_HTTP_HEADER", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"pcName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulNameLength": SimTypeInt(signed=False, label="UInt32"), "pcValue": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulValueLength": SimTypeInt(signed=False, label="UInt32")}, name="WEB_SOCKET_HTTP_HEADER", pack=False, align=None), offset=0), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hWebSocket", "pszSubprotocols", "ulSubprotocolCount", "pszExtensions", "ulExtensionCount", "pInitialHeaders", "ulInitialHeaderCount", "pAdditionalHeaders", "pulAdditionalHeaderCount"]),
        # 
        'WebSocketEndClientHandshake': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"pcName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulNameLength": SimTypeInt(signed=False, label="UInt32"), "pcValue": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulValueLength": SimTypeInt(signed=False, label="UInt32")}, name="WEB_SOCKET_HTTP_HEADER", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hWebSocket", "pResponseHeaders", "ulReponseHeaderCount", "pulSelectedExtensions", "pulSelectedExtensionCount", "pulSelectedSubprotocol"]),
        # 
        'WebSocketCreateServerHandle': SimTypeFunction([SimTypePointer(SimStruct({"Type": SimTypeInt(signed=False, label="WEB_SOCKET_PROPERTY_TYPE"), "pvValue": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "ulValueSize": SimTypeInt(signed=False, label="UInt32")}, name="WEB_SOCKET_PROPERTY", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pProperties", "ulPropertyCount", "phWebSocket"]),
        # 
        'WebSocketBeginServerHandshake': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"pcName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulNameLength": SimTypeInt(signed=False, label="UInt32"), "pcValue": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulValueLength": SimTypeInt(signed=False, label="UInt32")}, name="WEB_SOCKET_HTTP_HEADER", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"pcName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulNameLength": SimTypeInt(signed=False, label="UInt32"), "pcValue": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulValueLength": SimTypeInt(signed=False, label="UInt32")}, name="WEB_SOCKET_HTTP_HEADER", pack=False, align=None), offset=0), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hWebSocket", "pszSubprotocolSelected", "pszExtensionSelected", "ulExtensionSelectedCount", "pRequestHeaders", "ulRequestHeaderCount", "pResponseHeaders", "pulResponseHeaderCount"]),
        # 
        'WebSocketEndServerHandshake': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hWebSocket"]),
        # 
        'WebSocketSend': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="WEB_SOCKET_BUFFER_TYPE"), SimTypePointer(SimUnion({"Data": SimStruct({"pbBuffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulBufferLength": SimTypeInt(signed=False, label="UInt32")}, name="_Data_e__Struct", pack=False, align=None), "CloseStatus": SimStruct({"pbReason": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulReasonLength": SimTypeInt(signed=False, label="UInt32"), "usStatus": SimTypeShort(signed=False, label="UInt16")}, name="_CloseStatus_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hWebSocket", "BufferType", "pBuffer", "Context"]),
        # 
        'WebSocketReceive': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimUnion({"Data": SimStruct({"pbBuffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulBufferLength": SimTypeInt(signed=False, label="UInt32")}, name="_Data_e__Struct", pack=False, align=None), "CloseStatus": SimStruct({"pbReason": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulReasonLength": SimTypeInt(signed=False, label="UInt32"), "usStatus": SimTypeShort(signed=False, label="UInt16")}, name="_CloseStatus_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hWebSocket", "pBuffer", "pvContext"]),
        # 
        'WebSocketGetAction': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="WEB_SOCKET_ACTION_QUEUE"), SimTypePointer(SimUnion({"Data": SimStruct({"pbBuffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulBufferLength": SimTypeInt(signed=False, label="UInt32")}, name="_Data_e__Struct", pack=False, align=None), "CloseStatus": SimStruct({"pbReason": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulReasonLength": SimTypeInt(signed=False, label="UInt32"), "usStatus": SimTypeShort(signed=False, label="UInt16")}, name="_CloseStatus_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="WEB_SOCKET_ACTION"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="WEB_SOCKET_BUFFER_TYPE"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hWebSocket", "eActionQueue", "pDataBuffers", "pulDataBufferCount", "pAction", "pBufferType", "pvApplicationContext", "pvActionContext"]),
        # 
        'WebSocketCompleteAction': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["hWebSocket", "pvActionContext", "ulBytesTransferred"]),
        # 
        'WebSocketAbortHandle': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["hWebSocket"]),
        # 
        'WebSocketDeleteHandle': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["hWebSocket"]),
        # 
        'WebSocketGetGlobalProperty': SimTypeFunction([SimTypeInt(signed=False, label="WEB_SOCKET_PROPERTY_TYPE"), SimTypePointer(SimTypeBottom(label="Void"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["eType", "pvValue", "ulSize"]),
    }

lib.set_prototypes(prototypes)



import angr

######################################
# stub, for unsupported syscalls
######################################

#pylint:disable=redefined-builtin,arguments-differ
class syscall(angr.SimProcedure):

    def run(self, *args, resolves=None):

        self.resolves = resolves  # pylint:disable=attribute-defined-outside-init

        if self.successors is not None:
            self.successors.artifacts['resolves'] = resolves

        return self.state.solver.Unconstrained("syscall_stub_%s" % self.display_name, self.state.arch.bits, key=('syscall', '?', self.display_name))

    def __repr__(self):
        if 'resolves' in self.kwargs:
            return '<Syscall stub (%s)>' % self.kwargs['resolves']
        else:
            return '<Syscall stub>'



from .plugin import KnowledgeBasePlugin


class Propagations(KnowledgeBasePlugin):

    def __init__(self, kb):
        self._kb = kb
        self._propagations = {}

    def exists(self, func_loc):
        """
        Internal function to check if a func, specified as a CodeLocation
        exists in our known propagations

        :param func_loc:    CodeLocation of function
        :return:            Bool
        """
        exists = False
        for props in self._propagations:
            if props.block_addr == func_loc.block_addr:
                exists = True

        return exists

    def update(self, func_loc, replacements):
        """
        Add the replacements to known propagations

        :param func_loc:        CodeLocation of function
        :param replacements:    Dict of replacements
        """
        self._propagations[func_loc] = replacements

    def get(self, func_loc):
        """
        Gets the replacements for a specified function location.
        If the replacement does not exist in the known propagations, it
        returns None.

        :param func_loc:    CodeLocation of function
        :return:            Dict or None
        """
        if self.exists(func_loc):
            return self._propagations[func_loc]
        else:
            return None

    def copy(self):
        o = Propagations(self._kb)
        o._propagations = {}
        for k, v in self._propagations.items():
            o._propagations[k] = v


KnowledgeBasePlugin.register_default('propagations', Propagations)



# pylint:disable=unused-argument,arguments-differ
import ailment

from ..sequence_walker import SequenceWalker
from ..structurer_nodes import SequenceNode, CodeNode, MultiNode, LoopNode, ConditionNode, CascadingConditionNode


class CascadingIfsRemover(SequenceWalker):
    """
    Coalesce cascading If constructs. Transforming the following construct::

        if (cond_a) {
            if (cond_b) {
                true_body
            } else { }
        } else { }

    into::

        if (cond_a and cond_b) {
            true_body
        } else { }
    """
    def __init__(self, node):
        handlers = {
            SequenceNode: self._handle_Sequence,
            CodeNode: self._handle_Code,
            MultiNode: self._handle_MultiNode,
            LoopNode: self._handle_Loop,
            ConditionNode: self._handle_Condition,
            CascadingConditionNode: self._handle_CascadingCondition,
        }

        super().__init__(handlers)
        self.walk(node)

    def _handle_Condition(self, node, parent=None, index=None, **kwargs):
        """

        :param ConditionNode node:
        :param successor:
        :return:
        """

        if node.true_node is not None:
            self._handle(node.true_node, parent=node, index=0)
        if node.false_node is not None:
            self._handle(node.false_node, parent=node, index=1)

        if node.true_node is not None and node.false_node is None:
            if isinstance(node.true_node, SequenceNode):
                last_node = None
                if len(node.true_node.nodes) > 1 and \
                        all(self.is_empty_node(node_) for node_ in node.true_node.nodes[:-1]):
                    last_node = node.true_node.nodes[-1]
                elif len(node.true_node.nodes) == 1:
                    last_node = node.true_node.nodes[0]

                true_node = last_node

                if isinstance(true_node, ConditionNode) and \
                        true_node.true_node is not None and \
                        true_node.false_node is None:
                    node.condition = ailment.BinaryOp(None, "LogicalAnd", (node.condition, true_node.condition), False,
                                                      **node.condition.tags)
                    node.true_node = true_node.true_node

    @staticmethod
    def is_empty_node(node):
        if isinstance(node, ailment.Block):
            return not node.statements
        if isinstance(node, SequenceNode):
            return all(CascadingIfsRemover.is_empty_node(n) for n in node.nodes)
        return False



# pylint:disable=no-member
from ...protos import cfg_pb2
from ...serializable import Serializable


class MemoryDataSort:
    Unspecified = None
    Unknown = "unknown"
    Integer = "integer"
    PointerArray = "pointer-array"
    String = "string"
    UnicodeString = "unicode"
    SegmentBoundary = "segment-boundary"
    CodeReference = "code reference"
    GOTPLTEntry = "GOT PLT Entry"
    ELFHeader = 'elf-header'
    FloatingPoint = 'fp'  # the size is determined by the MemoryData itself


_SORT_TO_IDX = {
    MemoryDataSort.Unspecified: cfg_pb2.MemoryData.Unspecified,
    MemoryDataSort.Unknown: cfg_pb2.MemoryData.UnknownDataType,
    MemoryDataSort.Integer: cfg_pb2.MemoryData.Integer,
    MemoryDataSort.PointerArray: cfg_pb2.MemoryData.PointerArray,
    MemoryDataSort.String: cfg_pb2.MemoryData.String,
    MemoryDataSort.UnicodeString: cfg_pb2.MemoryData.UnicodeString,
    MemoryDataSort.SegmentBoundary: cfg_pb2.MemoryData.SegmentBoundary,
    MemoryDataSort.CodeReference: cfg_pb2.MemoryData.CodeReference,
    MemoryDataSort.GOTPLTEntry: cfg_pb2.MemoryData.GOTPLTEntry,
    MemoryDataSort.ELFHeader: cfg_pb2.MemoryData.ELFHeader,
    MemoryDataSort.FloatingPoint: cfg_pb2.MemoryData.FloatingPoint,
}

_IDX_TO_SORT = dict((v, k) for k, v in _SORT_TO_IDX.items())


class MemoryData(Serializable):
    """
    MemoryData describes the syntactic content of a single address of memory.
    """

    __slots__ = ('addr', 'size', 'sort', 'max_size', 'pointer_addr', 'content', )

    def __init__(self, address, size, sort, pointer_addr=None, max_size=None):
        self.addr = address
        self.size = size
        self.sort = sort

        self.max_size = max_size
        self.pointer_addr = pointer_addr

        self.content = None  # optional

    @property
    def address(self):
        return self.addr

    def __repr__(self):
        return "\\%#x, %s, %s/" % (self.address,
                                   "%d bytes" % self.size if self.size is not None else "size unknown",
                                   self.sort
                                   )

    def copy(self):
        """
        Make a copy of the MemoryData.

        :return: A copy of the MemoryData instance.
        :rtype: MemoryData
        """
        s = MemoryData(self.address, self.size, self.sort, pointer_addr=self.pointer_addr, max_size=self.max_size)
        s.content = self.content

        return s

    def fill_content(self, loader):
        """
        Load data to fill self.content.

        :param loader:  The project loader.
        :return:        None
        """

        if self.sort == MemoryDataSort.String:
            self.content = loader.memory.load(self.addr, self.size)
            if self.content.endswith(b"\x00"):
                self.content = self.content.strip(b"\x00")
        else:
            # FIXME: Other types are not supported yet
            return

    #
    # Serialization
    #

    @classmethod
    def _get_cmsg(cls):
        return cfg_pb2.MemoryData()

    def serialize_to_cmessage(self):
        cmsg = self._get_cmsg()
        cmsg.ea = self.addr
        cmsg.size = self.size if self.size is not None else 0
        cmsg.type = _SORT_TO_IDX[self.sort]
        return cmsg

    @classmethod
    def parse_from_cmessage(cls, cmsg, **kwargs):
        md = cls(cmsg.ea, cmsg.size, _IDX_TO_SORT[cmsg.type])
        return md



from typing import Dict, Set, TYPE_CHECKING
from collections import defaultdict

from ...code_location import CodeLocation

if TYPE_CHECKING:
    from .definition import Definition


class Uses:

    __slots__ = ('_uses_by_definition', '_uses_by_location' )

    def __init__(self):
        self._uses_by_definition: Dict['Definition',Set[CodeLocation]] = defaultdict(set)
        self._uses_by_location: Dict[CodeLocation, Set['Definition']] = defaultdict(set)

    def add_use(self, definition: "Definition", codeloc: CodeLocation):
        """
        Add a use for a given definition.

        :param angr.analyses.reaching_definitions.definition.Definition definition: The definition that is used.
        :param codeloc: The code location where the use occurs.
        """
        self._uses_by_definition[definition].add(codeloc)
        self._uses_by_location[codeloc].add(definition)

    def get_uses(self, definition: 'Definition') -> Set[CodeLocation]:
        """
        Retrieve the uses of a given definition.

        :param definition: The definition for which we get the uses.
        """
        return self._uses_by_definition.get(definition, set())

    def remove_use(self, definition: 'Definition', codeloc: 'CodeLocation') -> None:
        """
        Remove one use of a given definition.

        :param definition:  The definition of which to remove the uses.
        :param codeloc:     The code location where the use is.
        :return:            None
        """
        if definition in self._uses_by_definition:
            if codeloc in self._uses_by_definition[definition]:
                self._uses_by_definition[definition].remove(codeloc)

        if codeloc in self._uses_by_location:
            self._uses_by_location[codeloc].remove(definition)

    def remove_uses(self, definition: 'Definition'):
        """
        Remove all uses of a given definition.

        :param definition:  The definition of which to remove the uses.
        :return:            None
        """
        if definition in self._uses_by_definition:
            codelocs = self._uses_by_definition[definition]
            del self._uses_by_definition[definition]

            for codeloc in codelocs:
                self._uses_by_location[codeloc].remove(definition)

    def get_uses_by_location(self, codeloc: CodeLocation) -> Set:
        """
        Retrieve all definitions that are used at a given location.

        :param codeloc: The code location.
        :return:        A set of definitions that are used at the given location.
        """
        return self._uses_by_location.get(codeloc, set())

    def copy(self):
        """
        Copy the instance.

        :return angr.angr.analyses.reaching_definitions.uses.Uses: Return a new <Uses> instance containing the same data.
        """
        u = Uses()
        u._uses_by_definition = defaultdict(set, ((k, set(v)) for k, v in self._uses_by_definition.items()))
        u._uses_by_location = defaultdict(set, ((k, set(v)) for k, v in self._uses_by_location.items()))

        return u

    def merge(self, other) -> bool:
        """
        Merge an instance of <Uses> into the current instance.

        :param angr.angr.analyses.reaching_definitions.uses.Uses other: The other <Uses> from which the data will be added
                                                                        to the current instance.
        :return: True if any merge occurred, False otherwise
        """
        merge_occurred = False

        for k, v in other._uses_by_definition.items():
            if k not in self._uses_by_definition:
                self._uses_by_definition[k] = v
                merge_occurred = True
            elif not v.issubset(self._uses_by_definition[k]):
                merge_occurred = True
                self._uses_by_definition[k] |= v

        for k, v in other._uses_by_location.items():
            if k not in self._uses_by_location:
                self._uses_by_location[k] = v
                merge_occurred = True
            elif not v.issubset(self._uses_by_location[k]):
                merge_occurred = True
                self._uses_by_location[k] |= v

        return merge_occurred
